/*  BEGIN software license
 *
 *  msXpertSuite - mass spectrometry software suite
 *  -----------------------------------------------
 *  Copyright(C) 2009, 2017 Filippo Rusconi
 *
 *  http://www.msxpertsuite.org
 *
 *  This file is part of the msXpertSuite project.
 *
 *  The msXpertSuite project is the successor of the massXpert project. This
 *  project now includes various independent modules:
 *  
 *  - massXpert, model polymer chemistries and simulate mass spectrometric data;
 *  - mineXpert, a powerful TIC chromatogram/mass spectrum viewer/miner;
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * END software license
*/



#pragma once

/////////////////////// Qt includes
#include <QObject>
#include <QString>
#include <QWidget>
#include <QBrush>
#include <QColor>


/////////////////////// QCustomPlot
//#include "qcustomplot-fr.hpp"
#include "qcustomplot.h"

///////////////////////////// Local includes
#include "History.hpp"
#include "MassSpecDataSet.hpp"


namespace msXpSmineXpert
{


	class AbstractPlotWidget : public QCustomPlot
	{
		Q_OBJECT

		public:
			QWidget *mp_parentWnd = 0;

			QString m_name = "NOT_SET";
			Q_PROPERTY(QString m_name WRITE setName READ name);

			QString m_desc = "NOT_SET";
			Q_PROPERTY(QString m_desc WRITE setDesc READ desc);

			// This value is used when a Tic integration is asked in the various
			// derived classes. While the functions need to be defined in the derived
			// classes, this member datum can lie here.
			double m_lastI = qSNaN();

			// This pair of values are used to qualify the key-axis (X axis) range
			// selected for the the TIC intensity integration calculation.
			double m_keyRangeStart = qSNaN();
			double m_keyRangeEnd = qSNaN();

			// We do now own these data. Only when the derived TicChromPlotWidget get
			// that pointer set the caller does setParent() of that point that plot
			// widget such that when the widget is destroyed, the data set also.
			MassSpecDataSet *mp_massSpecDataSet = nullptr;

			bool m_isMultiGraph = false;

			QCPColorMap *mp_colorMap = nullptr;

			// The filename from which the mass data were read in the first place.
			QString m_fileName;

			History m_history;

			QMenu *m_contextMenu = nullptr;

			QCPItemRect *m_zoomRect   = nullptr;
			QCPItemLine *m_selectLine = nullptr;
			QCPItemText *m_descText   = nullptr;
			QCPItemText *m_xDeltaText = nullptr;

			// Handling of the tracers
			bool         m_tracersVisible = true;
			QCPItemLine *m_hStartTracer;
			QCPItemLine *m_vStartTracer;
			QCPItemLine *m_endTracer /*only vertical*/;

			bool m_clickWasOnXAxis = false;
			bool m_clickWasOnYAxis = false;

			int               m_lastAxisRangeHistoryIndex = -1;
			QList<QCPRange *> m_xAxisRangeHistory;
			QList<QCPRange *> m_yAxisRangeHistory;

			QPen m_pen;

			int m_mouseMoveHandlerSkipAmount = 10;
			int m_mouseMoveHandlerSkipCount  = 0;

			QList<int> m_regQtKeyCodeList;
			int        m_pressedKeyCode;
			bool       m_isDragging = false;

			QPointF m_startDragPoint;
			QPointF m_currentDragPoint;

			QPointF m_lastMousedPlotPoint;

			double m_xRangeMin = 0;
			double m_xRangeMax = 0;
			double m_xDelta    = 0;

			double m_yRangeMin = 0;
			double m_yRangeMax = 0;
			double m_yDelta    = 0;

			// QColor m_unfocusedColor = QColor(Qt::lightGray);
			// QColor m_unfocusedColor = QColor(230, 230, 230, 255);
			QColor m_unfocusedColor = QColor("lightgray");
			QBrush m_unfocusedBrush = QBrush(m_unfocusedColor);

			QColor m_focusedColor = QColor(Qt::transparent);
			QBrush m_focusedBrush = QBrush(m_focusedColor);

			virtual bool setupWidget();
			virtual void createContextMenu();

			void clearHistory();
			void addHistoryItem(HistoryItem *item);
			void showHistory();
			Q_INVOKABLE QString history();

			void clearAxisRangeHistory();
			void updateAxisRangeHistory();
			void backOneAxisHistoryStep();
			void setAxisRangesWithHistoryIndex(int index);

			void hideAllPlotItems();
			void showTracers();
			void hideTracers();

			bool isSelectionARectangle();
			void drawRectangleAndZoom();
			void drawXDeltaLineAndMeasure();
			void drawXDeltaLineForZoomOrIntegration();
			bool isClickOntoAnyAxis(const QPointF &mousePoint);
			bool isClickOntoXAxis(const QPointF &mousePoint);
			bool isClickOntoYAxis(const QPointF &mousePoint);
			void calculateSortedDragDeltas();
			void yMinMaxOnXAxisCurrentRange(double &min, double &max,
					QCPGraph *g = nullptr);
			void axisRescale();

			// Construction/Destruction

			explicit AbstractPlotWidget(QWidget *parent, QWidget *parentWnd, const QString &name,
					const QString &desc, MassSpecDataSet *massSpecDataSet,
					const QString &fileName = QString());

			virtual ~AbstractPlotWidget();

			virtual void registerQtKey(int qtKeyCode);
			virtual void clearPlot();

			virtual void mouseMoveHandler(QMouseEvent *event);
			virtual void selectionChangedHandler();
			virtual void mousePressHandler(QMouseEvent *event);
			virtual void mouseReleaseHandler(QMouseEvent *event);
			virtual void mouseReleaseHandledEvent(QMouseEvent *       event,
					AbstractPlotWidget *emitter) = 0;
			virtual void axisDoubleClickHandler(QCPAxis *               axis,
					QCPAxis::SelectablePart part,
					QMouseEvent *           event);

			virtual void keyPressEvent(QKeyEvent *event);
			virtual void keyReleaseEvent(QKeyEvent *event);

			void replotWithAxisRange(QCPRange xAxisRange, QCPRange yAxisRange,
					int whichAxis);
			Q_INVOKABLE void replotWithAxisRangeX(double lower, double upper);
			Q_INVOKABLE void replotWithAxisRangeY(double lower, double upper);

			double valueAtKeyInGraph(double k, QCPGraph *g = nullptr);
			Q_INVOKABLE double getYatX(double x);

			void redrawPlotBackground(QWidget *focusedPlotWidget);

			Q_INVOKABLE void shouldDestroyPlotWidget();

			void toggleMultiGraphGraph();
			void showMultiGraphGraph();
			void hideMultiGraphGraph();

			QColor color();
			void changeColor(const QColor& newColor);

			virtual void exportData();
			virtual void exportPlot();

			virtual IntegrationType arbitraryIntegrationType() = 0;

			public slots:
				void setName(QString name);
			QString name();

			void setDesc(QString desc);
			QString desc();


signals:
			void updateStatusBarSignal(QString msg);

			void destroyMatchingMultiGraphGraph(AbstractPlotWidget *widget);

			void aboutToDestroyPlotWidget(AbstractPlotWidget *widget = nullptr);
			void destroyPlotWidget(AbstractPlotWidget *widget = nullptr);

			void toggleMultiGraphGraphSignal(AbstractPlotWidget *widget,
					bool                recursive);

	};


} // namespace msXpSmineXpert


