/*  BEGIN software license
 *
 *  msXpertSuite - mass spectrometry software suite
 *  -----------------------------------------------
 *  Copyright(C) 2009, 2017 Filippo Rusconi
 *
 *  http://www.msxpertsuite.org
 *
 *  This file is part of the msXpertSuite project.
 *
 *  The msXpertSuite project is the successor of the massXpert project. This
 *  project now includes various independent modules:
 *  
 *  - massXpert, model polymer chemistries and simulate mass spectrometric data;
 *  - mineXpert, a powerful TIC chromatogram/mass spectrum viewer/miner;
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * END software license
*/



/////////////////////// Qt includes
#include <QLineEdit>


/////////////////////// Local includes
#include "ConsoleWnd.hpp"
#include "Application.hpp"


namespace msXpSmineXpert
{

	ConsoleWnd::ConsoleWnd(QWidget *parent, const QString &applicationName)
		: QMainWindow(parent), m_applicationName{applicationName}
	{
		if(parent == nullptr)
			qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

		m_ui.setupUi(this);

		// This Console window may be used by more than a single application, thus
		// set the app name along with the title.
		setWindowTitle(QString("%1 - Console").arg(m_applicationName));
		readSettings();

		show();
	}

	ConsoleWnd::~ConsoleWnd()
	{
		writeSettings();
	}


	void
		ConsoleWnd::closeEvent(QCloseEvent *event)
		{
			writeSettings();
			event->accept();
		}


	void
		ConsoleWnd::writeSettings()
		{
			QSettings settings;
			settings.beginGroup("ConsoleWnd");
			settings.setValue("geometry", saveGeometry());
			settings.setValue("windowState", saveState());
			settings.endGroup();
		}


	void
		ConsoleWnd::readSettings()
		{
			QSettings settings;
			settings.beginGroup("ConsoleWnd");
			restoreGeometry(settings.value("geometry").toByteArray());
			restoreState(settings.value("windowState").toByteArray());
			settings.endGroup();
		}


	void
		ConsoleWnd::logMessage(QString msg, const QColor &color, bool overwrite)
		{
			Application *application = static_cast<Application *>(qApp);
			application->processEvents();

			// If overwrite is true, the last non-empty text block of the console
			// window text edit document is replaced with msg. This is useful when
			// logging repetitive messages like when loading mzML/db files where the
			// feedback provides the number and index of spectra being read from file.
			// We do not want to send pages and pages of unuseful stuff.

			// We always append text, so get the last block of the document.

			QTextBlock lastTextBlock = m_ui.consolePlainTextEdit->document()->lastBlock();

			// Get the text in that text block:

			QString blockText = lastTextBlock.text();

			// We'll need a cursor to move inside that text block or to go up one text
			// block. The cursor is located at the beginning of the text.

			QTextCursor textCursor(lastTextBlock);

			// Now start doing the replacing of the text if overwrite is true.

			if(overwrite)
			{
				// We want to remove the text of the last text block. If the line is
				// empty, remove the previous block. The last block might be empty if
				// the last appended text contained a \n.

				if(blockText.isEmpty())
				{
					// We need to go up one block while maintaining the anchor, thus
					// making a selection of text that we'll then delete. The position is
					// at the beginning of the new text block, so we effectively select
					// all the text of the previous block.

					textCursor.movePosition(QTextCursor::PreviousBlock, QTextCursor::KeepAnchor);
				}
				else
				{
					// Go to the end of the text block, keeping anchor, so as to select
					// the text in the last block. We'll remove it right after having
					// selected it.

					textCursor.movePosition(QTextCursor::EndOfBlock, QTextCursor::KeepAnchor);
				}

				// Finally, whatever the block movements, remove the selected text.

				textCursor.removeSelectedText();
			}

			// Now that we may have removed the last text block text, 
			// we can simply append the new msg text.

			// If the color is valid, then use that color to show the text.
			if(color.isValid())
			{
				QTextCharFormat prevCharFormat =
					m_ui.consolePlainTextEdit->currentCharFormat();
				QTextCharFormat newCharFormat = prevCharFormat;

				newCharFormat.setForeground(QBrush(color));

				m_ui.consolePlainTextEdit->setCurrentCharFormat(newCharFormat);

				m_ui.consolePlainTextEdit->appendPlainText(msg);
				m_ui.consolePlainTextEdit->setCurrentCharFormat(prevCharFormat);
			}
			else
			{
				QTextCharFormat prevCharFormat =
					m_ui.consolePlainTextEdit->currentCharFormat();
				prevCharFormat.setForeground(QBrush(QColor("black")));
				m_ui.consolePlainTextEdit->setCurrentCharFormat(prevCharFormat);

				m_ui.consolePlainTextEdit->appendPlainText(msg);
			}

			// Make sure that the document scrolls.
			QTextCursor cursor = m_ui.consolePlainTextEdit->textCursor();
			cursor.movePosition(QTextCursor::End);
			m_ui.consolePlainTextEdit->setTextCursor(cursor);
		};


	// Overloaded without color for scripting simplicity.
	void
		ConsoleWnd::logMessage(QString msg)
		{
			logMessage(msg, QColor());
		}

} // namespace msXpSmineXpert
