/*  BEGIN software license
 *
 *  msXpertSuite - mass spectrometry software suite
 *  -----------------------------------------------
 *  Copyright(C) 2009, 2017 Filippo Rusconi
 *
 *  http://www.msxpertsuite.org
 *
 *  This file is part of the msXpertSuite project.
 *
 *  The msXpertSuite project is the successor of the massXpert project. This
 *  project now includes various independent modules:
 *  
 *  - massXpert, model polymer chemistries and simulate mass spectrometric data;
 *  - mineXpert, a powerful TIC chromatogram/mass spectrum viewer/miner;
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * END software license
*/



#ifndef MAIN_WINDOW_HPP
#define MAIN_WINDOW_HPP


/////////////////////// Qt includes
#include <QObject>
#include <QMainWindow>
#include <QHash>
#include <QMap>


/////////////////////// Local includes
#include "globals.hpp"

#include "AnalysisPreferences.hpp"
#include "AnalysisPreferencesDlg.hpp"

#include "AbstractPlotWidget.hpp"

#include "ConsoleWnd.hpp"
#include "libmassgui/ScriptingWnd.hpp"
#include "TicChromWnd.hpp"
#include "DriftSpecWnd.hpp"
#include "MassSpecWnd.hpp"
#include "ColorMapWnd.hpp"

#include "MassSpecDataSet.hpp"
#include "OpenSpectraDlg.hpp"

#include "MassSpecDataFileLoaderSqlite3.hpp"

#include "DataPlotWidgetRelation.hpp"


namespace msXpSmineXpert
{

	class MainWindow : public QMainWindow
	{
		Q_OBJECT

    private:
    QString m_moduleName;

		public:

			// This is the object that will be responsible for maintaining all the
			// relations between various plot widgets, their data set and the open
			// spectra dialog list widget's item. It belongs here because it needs to
			// be accessible from *anywhere* in the program.
			DataPlotWidgetRelationer m_dataPlotWidgetRelationer;

			// The dialog window in which the user specifies the file for the
			// recording of the analysis stanzas and the characteristics of
			// the mass spec experiments such that the stanzas are informative.
			AnalysisPreferencesDlg *mp_analysisPrefDlg = nullptr;

			// The window in which the Total Ion Current chromatogram is to be
			// displayed. The widget in which the chromatogram is displayed must be
			// known in *this window, because it will be asked for by the Mass Spec
			// window. Each time a new mass data file is successfully opened, a
			// MassSpecDataSet *massSpecDataSet object is allocated to store the data
			// that will serve to the computation of the TIC chromatogram. That
			// MassSpecDataSet instance needs to be stored along with the widget in
			// which the TIC chromatogram is displayed (see below).
			TicChromWnd *mp_ticChromWnd = nullptr;

			// The window in which the mass spectrum is displayed. The mass spectrum
			// is computed by combining mass spectra as triggered by dragging the
			// mouse on one of the the TicChromWnd-contained plot widgets.
			MassSpecWnd *mp_massSpecWnd = nullptr;

			// The window in which the drift spectrum is displayed. The drift spectrum
			// is computed by combining mass spectra as triggered by dragging the
			// mouse on one of the the TicChromWnd-contained or MassSpecWnd-contained
			// plot widgets.
			DriftSpecWnd *mp_driftSpecWnd = nullptr;

			// The window in which the color map is displayed. The color map will
			// display with an artificial colors map the dt=f(mz) or vice-versa data.
			ColorMapWnd *mp_colorMapWnd = nullptr;

			// The window in which important messages are displayed. This window must
			// be accessible from anywhere in the program.
			ConsoleWnd *mp_consoleWnd = nullptr;

			// The scripting window. This window must be accessible from 
			// anywhere in the program.
			msXpSlibmassgui::ScriptingWnd *mp_scriptingWnd = nullptr;

			// Each time a new mass spec file is opened, a corresponding plot widget is
			// automatically created to plot the TIC chromatogram. This task is the
			// same whatever the kind of mass spec data file.
			void setupTicChromPlotWidget(MassSpecDataSet *massSpecDataSet);

			// The window in which a list widget will show all the opened mass files.
			// Each time a new mass data file is opened, a new item in this window is
			// shown, the item of the list widget that it contains being the filename
			// of the mass data file. It is important that there be a connection
			// between the list widget item and the corresponding TicChromWnd plot
			// widget that displays the corresonding TIC chromatogram.
			OpenSpectraDlg *mp_openSpectraDlg = nullptr;

			// This is the progress bar that is shown in the status bar.
			QProgressBar *mp_statusBarProgressBar = nullptr;

			// When launching lengthy operations, let the user cancell them if
			// necessary.
			QPushButton *mp_statusBarCancelPushButton = nullptr;

			// This is the label that is shown in the status bar.
			QLabel *mp_statusBarLabel = nullptr;

			bool m_isOperationCancelled = false;

			AnalysisPreferences *mpa_analysisPreferences = nullptr;
			QFile                m_analysisFile;

			QStringList m_nameFilters = {"All files (*.* *)"};

			QList<QColor *> m_colorList;
			QList<QColor *> m_usedColorList;
			int             m_lastUsedColorIndex = -1;

			double m_mzSpectrumStart;
			double m_mzSpectrumEnd;

			double m_mzRangeStart;
			double m_mzRangeEnd;

			QMenu *m_fileMenu;
			QMenu *m_plotMenu;
			QMenu *m_windowsMenu;
			QMenu *m_helpMenu;

			QAction *m_openFullMsFileAct;
			QAction *m_openStreamedMsFileAct;
			QAction *m_openMsClipboardAct;

			QAction *m_analysisPrefsAct;
			QAction *m_quitAct;
			QAction *m_clearPlotsAct;
			QAction *m_aboutAct;
			QAction *m_aboutQtAct;
			QAction *m_showTicChromWndAct;
			QAction *m_showMassSpecWndAct;
			QAction *m_showDriftSpecWndAct;
			QAction *m_showColorMapWndAct;
			QAction *m_showOpenSpectraDlgAct;
			QAction *m_showConsoleWndAct;
			QAction *m_showScriptingWndAct;
			QAction *m_showAllWindowsAct;
			QAction *m_saveWorkspaceAct;

			// Construction/destruction
    MainWindow(const QString &moduleName);
			~MainWindow();

			void createMenusAndActions();

			void writeSettings();
			void readSettings();

			Q_INVOKABLE void about();
			Q_INVOKABLE void quit();
			void clearPlots();

			void closeEvent(QCloseEvent *event);

			bool setupWindow();

			bool initialize();

			void initializeScripting();

			void recyclePlottingColor(QColor color);
			bool isPlottingColorUsed(const QColor &color) const;

			void loadChromatogram(
					const QString & = QString("total ion current chromatogram"));

			// fixme
			// void loadSqlData(MassSpecDataSet *massSpecDataSet);

			// If overwrite is true, then the last text block that is non-empty is
			// replaced with the msg text.
			void logConsoleMessage(QString msg, const QColor &color = QColor(), bool overwrite = false);

			void dataLoadMessage(QString message, int logType);

			void setupProgressFeedback(QString msg, int startValue, int endValue,
					int currentValue = -1);
			void updateProgressFeedback(QString msg, int currentValue);
			void statusBarMessage(QString msg);
			void statusBarProgress(int value);

			Q_INVOKABLE void createMassSpectrumClipboard();

			Q_INVOKABLE void openMassSpectrometryFile(const QString &fileName = QString(), bool isStreamed = false);
			Q_INVOKABLE void openMassSpectrometryFiles(const QStringList &fileList,
					bool isStreamed);

			void openMassSpectrometryFileSqlite3(QString fileName, bool isStreamed = false);
			// void openMassSpectrometryFileMzml(QString fileName, bool isStreamed = false);
			void openMassSpectrometryFilePwiz(QString fileName, bool isStreamed = false);
			void openMassSpectrometryFileDx(QString fileName, bool isStreamed = false);
			void openMassSpectrometryFileMs1(QString fileName, bool isStreamed = false);
			void openMassSpectrometryFileBrukerXy(QString fileName, bool isStreamed = false);
			void openMassSpectrometryFileXy(QString fileName);

			void makeLoaderConnections(MassSpecDataFileLoader *loader);

			void analysisPrefDlg();
			void showTicChromWnd();
			void showMassSpecWnd();
			void showDriftSpecWnd();
			void showColorMapWnd();
			void showOpenMassSpectraDlg();
			void showConsoleWnd();
			void showScriptingWnd();
			void showAllWindows();
			void saveWorkspace();


			void destroyAlsoAllTargetPlotWidgets(AbstractPlotWidget *originPlotWidget);

			// Multigraph graph visibility toggling.
			void toggleMultiGraphGraph(AbstractPlotWidget *widget, bool recursive);

			void cancelPushButtonClicked();

			public slots:
				void openFullMassSpectrometryFileDlg();
			void openStreamedMassSpectrometryFileDlg();


signals:
			void cancelOperationSignal();
	};

} // namespace msXpSmineXpert


#endif /* MAIN_WINDOW_HPP */
