/*  BEGIN software license
 *
 *  msXpertSuite - mass spectrometry software suite
 *  -----------------------------------------------
 *  Copyright(C) 2009, 2017 Filippo Rusconi
 *
 *  http://www.msxpertsuite.org
 *
 *  This file is part of the msXpertSuite project.
 *
 *  The msXpertSuite project is the successor of the massXpert project. This
 *  project now includes various independent modules:
 *  
 *  - massXpert, model polymer chemistries and simulate mass spectrometric data;
 *  - mineXpert, a powerful TIC chromatogram/mass spectrum viewer/miner;
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * END software license
*/



#ifndef MASS_SPEC_PLOT_WIDGET_HPP
#define MASS_SPEC_PLOT_WIDGET_HPP

/////////////////////// Qt includes
#include <QObject>
#include <QString>
#include <QWidget>
#include <QColor>


///////////////////////////// Local includes
#include "AbstractPlotWidget.hpp"
#include "AbstractMultiPlotWnd.hpp"
#include "MassSpecDataSet.hpp"
#include "History.hpp"


namespace msXpSmineXpert
{


	class MassSpecPlotWidget : public AbstractPlotWidget
	{
		Q_OBJECT

		public:
			// These data are required to display the charge and mass values as
			// calculated when the user drags the mouse over peaks belonging to either
			// an isotopic cluster or two peaks of the same charge envelope.
			int m_lastCharge = -1;
			double m_protonMass = 1.007825035;
			double m_lastMz = qSNaN();
			double m_lastMass = qSNaN();

			// Ideally that would be set to 1, but the graphical display of the mass
			// spec is pixel-based, and depending on the screen real estate, from a
			// pixel to the other, the m/z gap might be too much and then we would not
			// be able to ever show the deconvolution results. Setting this value too
			// low produces hysteresis, in the sense that by pointing exactly the same
			// points, but from left to right or right to left, the Mr values differ
			// significantly.
			double m_chargeFracPartTolerance = 0.99;

			bool deconvolute();
			bool deconvoluteIsotopicCluster();
			bool deconvoluteChargedState(int span = 1);

			IntegrationType arbitraryIntegrationType() override;

			QList<QCPGraph *> graphs();
			QList<QCPGraph *> unambiguousGraphs();

			bool isGraphSelected(QCPGraph *graph);
			QCPGraph *firstSelectedGraph();

			QString craftAnalysisStanza(QCPGraph *graph = nullptr, const QString &fileName = QString());

			// Construction/Desctruction
			MassSpecPlotWidget(QWidget *parent, AbstractMultiPlotWnd *parentWnd,
					const QString &name, const QString &desc,
					const QString & fileName = QString(),
					MassSpecDataSet *massSpecDataSet = nullptr);

			~MassSpecPlotWidget();

			void mouseMoveHandler(QMouseEvent *event) override;
			void mouseReleaseHandledEvent(QMouseEvent * event,
					AbstractPlotWidget *emitter) override;

			void keyReleaseEvent(QKeyEvent *event) override;

			Q_INVOKABLE void integrateToDt(double lower, double upper);
			Q_INVOKABLE void integrateToRt(double lower, double upper);
			Q_INVOKABLE double integrateToTicIntensity(double lower, double upper);

signals:
			void recordAnalysisStanza(QString stanza, const QColor &color = QColor());

			void newDriftSpectrumByMz(MassSpecDataSet *massSpecDataSet,
					IntegrationType integrationType,
					const History &history, QColor color);

			void newTicChromatogramByMz(MassSpecDataSet *massSpecDataSet,
					IntegrationType integrationType,
					const History &history, QColor color);
	};


} // namespace msXpSmineXpert


#endif // MASS_SPEC_PLOT_WIDGET_HPP
