/*  BEGIN software license
 *
 *  msXpertSuite - mass spectrometry software suite
 *  -----------------------------------------------
 *  Copyright(C) 2009, 2017 Filippo Rusconi
 *
 *  http://www.msxpertsuite.org
 *
 *  This file is part of the msXpertSuite project.
 *
 *  The msXpertSuite project is the successor of the massXpert project. This
 *  project now includes various independent modules:
 *  
 *  - massXpert, model polymer chemistries and simulate mass spectrometric data;
 *  - mineXpert, a powerful TIC chromatogram/mass spectrum viewer/miner;
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * END software license
*/



#include <qmath.h>
#include <omp.h>


/////////////////////// Qt includes
#include <QWidget>
#include <QClipboard>


/////////////////////// Local includes
#include "AbstractPlotWidget.hpp"
#include "MassSpecWnd.hpp"
#include "MainWindow.hpp"
#include "MassSpecDataFileLoaderPwiz.hpp"
#include "MassSpecDataFileLoaderDx.hpp"
#include "MassSpecDataFileLoaderMs1.hpp"


namespace msXpSmineXpert
{

	int MassSpecWnd::widgetCounter = 0;

	MassSpecWnd::MassSpecWnd(QWidget *parent)
		: AbstractMultiPlotWnd{parent, "mineXpert" /*m_name*/,
			"Mass-spectrum" /*m_desc*/}
	{
		initialize();

		show();
	}

	MassSpecWnd::~MassSpecWnd()
	{
	}

	void
		MassSpecWnd::initialize()
		{
			initializePlotRegion();

			// Now insert the permanent spin box widget at the far right of the status
			// bar where the span between two charge envelope peaks can be specified
			// to help compute the charge of a given mass peak and the molecular mass
			// of the analyte.
			mp_chargeSpanSpinBox = new QSpinBox(this);
			mp_chargeSpanSpinBox->setRange(1, 500);
			mp_chargeSpanSpinBox->setSingleStep(1);
			mp_chargeSpanSpinBox->setToolTip("Set this to the number of "
					"charge\nenvelope peaks that are spanned "
					"by\nthe mouse drag connection");

			// This widget will be appended to the others already set up in the base
			// class.
			statusBar()->addPermanentWidget(mp_chargeSpanSpinBox);
		}

	void
		MassSpecWnd::initializePlotRegion()
		{
			// There are two regions in the window. The upper part of the mp_splitter
			// is gonna receive the multi-graph plotwidget. The lower part of the
			// splitter will receive all the individual plot widgets.

			mp_multiGraphPlotWidget =
				new MassSpecPlotWidget(this, this, m_name, m_desc);
			mp_multiGraphPlotWidget->setObjectName("multiGraphPlot");

			// We will need later to know that this plot widget is actually for
			// displaying many different graphs, contrary to the normal situation.
			mp_multiGraphPlotWidget->m_isMultiGraph = true;

			// We need to remove the single graph that was created upon creation of
			// the QCustomPlot widget because we'll later add graphs, and we want to
			// have a proper correlation between the index of the graphs and their
			// pointer value. If we did not remove that graph "by default", then there
			// would be a gap of 1 between the index of the first added mass graph and
			// its expected value.
			mp_multiGraphPlotWidget->clearGraphs();

			// Ensure that there is a connection between this plot widget and this
			// containing window, such that if there is a requirement to replot, the
			// widget gets the signal.
			connect(this, &AbstractMultiPlotWnd::replotWithAxisRange,
					mp_multiGraphPlotWidget, &AbstractPlotWidget::replotWithAxisRange);

			// This window is responsible for the writing of the analysis stanzas to
			// the file. We need to catch the signals.
			connect(mp_multiGraphPlotWidget, &MassSpecPlotWidget::recordAnalysisStanza,
					this, &MassSpecWnd::recordAnalysisStanza);

			// This window is responsible for displaying the key/value pairs of the
			// graphs onto which the mouse is moved. We need to catch the signal.
			connect(mp_multiGraphPlotWidget, &AbstractPlotWidget::updateStatusBarSignal, this,
					&AbstractMultiPlotWnd::updateStatusBar);

			// We want to create a scroll area in which the multigraph plot widget
			// will be sitting.
			QScrollArea *scrollArea = new QScrollArea();
			scrollArea->setObjectName(QStringLiteral("scrollArea"));
			scrollArea->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOn);
			scrollArea->setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOn);
			// We accept that the plots are resized, but we set a minimum size when we
			// create them.
			scrollArea->setWidgetResizable(true);
			scrollArea->setAlignment(Qt::AlignCenter);
			scrollArea->setWidget(mp_multiGraphPlotWidget);
			// And now prepend to the mp_splitter that new scroll area. Remember that
			// the mp_splitter had already a widget set to it, that is, the scroll
			// area that is now located below the mp_multiGraphPlotWidget, that is
			// responsible for the hosting of all the single-graph plot widgets.
			mp_splitter->insertWidget(0, scrollArea);
		}


	bool
		MassSpecWnd::focusNextPrevChild(bool next)
		{
			// In this window, we do not want that the TAB key serves to move from
			// widget
			// to widget.
			return false;
		}


	MassSpecPlotWidget *
		MassSpecWnd::addPlotWidget(QVector<double> keyVector,
				QVector<double> valVector, const QString &desc,
				MassSpecDataSet *massSpecDataSet)
		{
			MainWindow *mainWindow = static_cast<MainWindow *>(parent());

			// Beware that massSpecDataSet might be nullptr.
			QString fileName;
			if(massSpecDataSet != nullptr)
				fileName = massSpecDataSet->fileName();
			else
				fileName = "NOT_SET";

			MassSpecPlotWidget *widget = new MassSpecPlotWidget(
					this, this, m_name, desc, fileName, massSpecDataSet);
			widget->setObjectName("monoGraphPlot");


			///////////////// BEGIN Script availability of the new widget //////////////////

			// Craft a name that differentiates this widget from the previous and next
			// ones. We use the static counter number to affix to the string
			// "massSpec".

			QString propName =
				QString("massSpecPlot%1").arg(MassSpecWnd::widgetCounter);
			QString alias = "lastMassSpecPlot";

			QString explanation = QString("Make available a mass spectrum plot under name \"%1\" "
					"with alias \"%2\".\n").arg(propName).arg(alias);

			// Make the new plot widget globally available under the crafted property
			// name and alias. Also, put out the explanation text. Note that by
			// providing 'this', we make the new object appear in the scripting objet
			// tree widget as a child item of this.

			QScriptValue scriptValue;
			mainWindow->mp_scriptingWnd->publishQObject(widget, this /* parent object */, 
					propName, alias, explanation, QString() /* help string */, scriptValue);

			++MassSpecWnd::widgetCounter;

			///////////////// END Script availability of the new widget //////////////////


			widget->setMinimumSize(400, 150);
			mp_scrollAreaVBoxLayout->addWidget(widget);

			widget->graph()->setData(keyVector, valVector);

			// Let's see if one color was set as the desired color, otherwise use
			// black.
			QPen currentPen = widget->graph()->pen();

			if(!m_wishedColor.isValid())
				currentPen.setColor(Qt::black);
			else
				currentPen.setColor(m_wishedColor);

			widget->graph()->setPen(currentPen);

			widget->rescaleAxes(true /*only enlarge*/);
			widget->replot();

			// Store in the history the new ranges of both x and y axes.
			widget->updateAxisRangeHistory();

			// At this point, make sure that the same data are represented on the
			// multi-graph plot widget.

			QCPGraph *newGraph = mp_multiGraphPlotWidget->addGraph();
			newGraph->setData(keyVector, valVector);
			newGraph->setPen(currentPen);
			mp_multiGraphPlotWidget->rescaleAxes(true /*only enlarge*/);
			mp_multiGraphPlotWidget->replot();

			// Maintain a hash connection between any given plot widget and the
			// corresponding graph in the multigraph widget.
			m_plotWidgetGraphMap.insert(widget, newGraph);


			// Finally append the new widget to the list of widgets that belong to
			// this window.
			m_plotWidgetList.append(widget);

			// Signal/slot connections.

			connect(widget, &MassSpecPlotWidget::recordAnalysisStanza, this,
					&MassSpecWnd::recordAnalysisStanza);

			connect(widget, &AbstractPlotWidget::updateStatusBarSignal, this,
					&AbstractMultiPlotWnd::updateStatusBar);

			return widget;
		}

	void
		MassSpecWnd::setAnalysisPreferences(AnalysisPreferences *preferences)
		{
			if(preferences == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			mp_analysisPreferences = preferences;
		}

	void
		MassSpecWnd::setAnalysisFile(QFile *analysisFile)
		{
			if(analysisFile == nullptr)
				qFatal(
						"Fatal error at %s@%d. The file pointer cannot be 0. Program aborted.",
						__FILE__, __LINE__);

			mp_analysisFile = analysisFile;
		}


	void
		MassSpecWnd::newMassSpectrumByRt(MassSpecDataSet *massSpecDataSet,
				IntegrationType integrationType,
				History history, QColor color)
		{
			if(massSpecDataSet == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			m_wishedColor = color;

			// We get a History reference that informs us about the ranges in the
			// sense that the innermost range for integrationType is by necessity the
			// integration range we work on.

			double rangeStart;
			double rangeEnd;

			if(!history.innermostRange(integrationType, &rangeStart, &rangeEnd))
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// qDebug() << __FILE__ << __LINE__
			//<< "rangeStart:" << rangeStart << "rangeEnd:" << rangeEnd;

			MainWindow *mainWindow = static_cast<MainWindow *>(parent());
			QString msgSkeleton =
				QString("Computing mass spectrum between rt %1 and rt %2")
				.arg(rangeStart, 0, 'f', 6)
				.arg(rangeEnd, 0, 'f', 6);
			mainWindow->logConsoleMessage(msgSkeleton);

#if 0
			// No more of use here at this is performed in the calling function.

			History localHistory = history;

			HistoryItem *histItem = new HistoryItem;
			histItem->newIntegrationRange(IntegrationType::RT_TO_MZ, rtStart, rtEnd);

			// Aggreate the new history item to the history we got as a parameter and
			// that we copied into localHistory.

			localHistory.appendHistoryItem(histItem);
#endif

			// Make sure the cancel button is visible.
			mp_statusBarCancelPushButton->setVisible(true);

			double start = omp_get_wtime();

			QVector<double> keyVector;
			QVector<double> valVector;

			if(massSpecDataSet->isStreamed())
			{
				newStreamedMassSpectrum(massSpecDataSet, integrationType, &keyVector,
						&valVector, history, color);
			}
			else // that is, !if(massSpecDataSet->isStreamed())
			{

				// At this point, we already know that we have to perform the work on
				// the basis of data that are in the RAM already.

				// We are doing an integration RT_TO-MZ, that is that we are selecting
				// retention times as defined by a retention time range :
				// [rtStart-rtEnd]. This is fine, but where are these data from ? Did
				// these data come from a previous integration or did they come right
				// over the initial RT ? Further, was there a single previous
				// integration or were there
				// multiple ones ? And what integrations were these ? DT-based ? Or
				// what ?

				// Let's see an example : start with the initial RT. Make a first
				// integration to MZ with a RT range. Then, in the MZ spectrum, make an
				// integration to DT with a MZ range. Finally, in the DT spectrum, make
				// an integration to MZ with a DT range. We need to always document
				// what we do and check each time all the previous integrations so as
				// to provide
				// the expected result.

				// MZ integration.
				// ===============
				// Already set at the top of this function.

				// MZ integration.
				// ===============
				bool mzIntegration = false;
				double mzStart = -1;
				double mzEnd = -1;

				// If the history is comprised of many successive integrations on the
				// basis of m/z value ranges, then we should make sure we do take into
				// account the more internal m/z value range.
				mzIntegration = history.innermostMzRange(&mzStart, &mzEnd);
				if(mzIntegration)
				{
					mainWindow->logConsoleMessage(
							"m/z integration accounted for: \n" +
							QString("[%1-%2]\n").arg(mzStart).arg(mzEnd));
				}

				// DT integration.
				// ===============
				bool dtIntegration = false;
				double dtStart = -1;
				double dtEnd = -1;

				// If the history is comprised of many successive integrations on the
				// basis of dt value ranges, then we should make sure we do take into
				// account the more internal dt value range.
				dtIntegration = history.innermostDtRange(&dtStart, &dtEnd);
				if(dtIntegration)
				{
					mainWindow->logConsoleMessage(
							"dt integration accounted for: \n" +
							QString("[%1-%2]\n").arg(dtStart).arg(dtEnd));
				}

				// Get handles to the original data in the calling entity.
				const msXpSlibmass::MsMultiHash &rtHash = massSpecDataSet->rtHash();

				// Because the main axis of integration is the retention time, get a
				// list of all the retention times that are stored in the spec data set
				// and that are contained in the rt range passed as parameters.

				QList<double> rtList;
				rtHash.rangedKeys(rtList, rangeStart, rangeEnd);

				qSort(rtList.begin(), rtList.end());

				int rtListSize = rtList.size();

				// qDebug() << __FILE__ << __LINE__
				// << "Number of unique retention time keys: " << rtListSize
				// << "from scan start time "<< rtList.first()
				// << "to scan start time" << rtList.last();

				setupProgressFeedback(msgSkeleton, 1, rtListSize);

				// We are to produce a mass spectrum that is the result of combinations
				// of spectra, so allocate one on the stack (quick calcs).

				msXpSlibmass::MassSpectrum combinedMassSpectrum;

				int iterationCount = 0;

#pragma omp parallel for
				for(int iter = 0; iter < rtListSize; ++iter)
				{

					if(m_isOperationCancelled)
						continue;

					double rt = rtList.at(iter);

					// qDebug() << __FILE__ << __LINE__ << "integrating for rt:" << rt;

					msXpSlibmass::MassSpectrumList rtMassSpectra(rtHash.values(rt));

					if(dtIntegration)
					{
						// We have to remove from all the rtMassSpectra, all the spectra
						// that
						// are not in the drift time range.

						for(int jter = 0; jter < rtMassSpectra.size(); ++jter)
						{
							msXpSlibmass::MassSpectrum *spectrum = rtMassSpectra.at(jter);

							// Old version when msXpSlibmass::MassSpectrum did not have m_rt/m_dt
							// member
							// variables.
							// double dt = dtHash.key(spectrum);

							if(spectrum->dt() < dtStart || spectrum->dt() > dtEnd)
							{
								// Remove the mass spectrum from the rtMassSpectra
								// because it does
								// not have the proper dt value.
								rtMassSpectra.removeAt(jter);
								--jter;
								continue;
							}
						}
					}

					// At this point, the rtMassSpectra only contain spectra that match
					// both the retention time range and the drift time range. Now,
					// perform the spectra combination by taking into account an
					// eventual
					// MZ-based integration.

					// qDebug() << __FILE__ << __LINE__
					// << "For rt " << rt << "the number of spectra is" <<
					// rtMassSpectra.size();

					QString msg = msgSkeleton +
						QString("-- rt %1 with %2 spectra")
						.arg(rt, 0, 'f', 6)
						.arg(rtMassSpectra.size());

#pragma omp critical
					{
						if(mzIntegration)
						{

							// We need to use a mz-specific combine function because each
							// MassPeak must be checked for its mz value that must be in
							// between mzStart and mzEnd to be combined.

							combinedMassSpectrum.combine(rtMassSpectra, mzStart, mzEnd);
						}
						else
						{
							combinedMassSpectrum.combine(rtMassSpectra);

#if 0
							// Debug stuff
							qDebug() << __FILE__ << __LINE__
								<< "Now outputting debug data to /tmp";

							for(int jter = 0; jter < rtMassSpectra.size(); ++jter)
							{
								QString fileName = QString("/tmp/test-%1-%2")
									.arg(iterationCount)
									.arg(jter);

								msXpSlibmass::MassSpectrum *massSpectrum = rtMassSpectra.at(jter);

								massSpectrum->debugAsText(fileName);
							}

							combinedMassSpectrum.debugAsText(QString("/tmp/combined-%1")
									.arg(iterationCount));
#endif
						}

						++iterationCount;

						// Debugging the odd combined spectra for Bruker QTOF files

#if 0
						// Debug stuff

						qDebug() << __FILE__ << __LINE__
							<< "Now outputting debug data to /tmp";

						QFile file(QString("/tmp/test-%1").arg(iterationCount));

						if(!file.open(QFile::WriteOnly | QFile::Truncate))
							qFatal("Fatal error at %s@%d. Program aborted.",
									__FILE__, __LINE__);

						QTextStream stream(&file);

						for(int iter = 0; iter < combinedMassSpectrum.size(); ++iter)
						{
							QString dataString = QString().setNum(combinedMassSpectrum.at(iter)->m_mz, 'f', 10);
							dataString += " ";
							dataString += QString().setNum(combinedMassSpectrum.at(iter)->m_i, 'f', 10);
							dataString += "\n";
							stream << dataString;

							qDebug() << __FILE__ << __LINE__ << dataString;
						}

						stream.flush();
						file.close();
#endif

						// The label and progress bar widgets were setup by the master
						// thread, so we can only modify them (actually emitting signals)
						// from that same master thread:

						if(omp_get_thread_num() == 0)
							updateProgressFeedback(msg, iterationCount);

					}
					// End of
					// #pragma omp critical

				}
				// End of
				// for(int iter = 0; iter < rtListSize; ++iter)

				keyVector = QVector<double>::fromList(combinedMassSpectrum.mzList());
				valVector = QVector<double>::fromList(combinedMassSpectrum.iList());
			}
			// End of
			// else // that is, !if(massSpecDataSet->isStreamed())

			double end = omp_get_wtime();

			QString msg = QString("Computed mass spectrum in %1 s").arg(end - start);

			// Show 100% completion of the work.
			setupProgressFeedback(msg, 0, 100, 100);

			// Hide the cancel button.
			mp_statusBarCancelPushButton->setVisible(false);

			mainWindow->logConsoleMessage(msg);

			// qDebug() << __FILE__ << __LINE__ << msg;

			// In some seconds, hide the progress bar and set the label to an empty
			// text.
			QTimer::singleShot(3000, this, [=]() {
					this->updateProgressFeedback("", 0);
					this->mp_statusBarProgressBar->setVisible(false);
					});

			// Sanity check
			if(keyVector.size() != valVector.size())
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// We may have gotten here because the operation was cancelled. In any case,
			// reset the boolean value.

			m_isOperationCancelled = false;

			// We need to plot that stuff into a new widget.

			// Create a text item to indicate the range.

			QString text = QString("%1\n%2 - rt range : [%3-%4] ")
				.arg(massSpecDataSet->fileName())
				.arg(m_desc)
				.arg(rangeStart, 0, 'f', 6)
				.arg(rangeEnd, 0, 'f', 6);

			// Allocate the new plot widget in which the data are to be plot.

			MassSpecPlotWidget *widget = static_cast<MassSpecPlotWidget *>(
					addPlotWidget(keyVector, valVector, text, massSpecDataSet));

			// The history that we get as a parameter needs to be appended to the new
			// widget's history so that we have a full account of the history.

			widget->m_history.copyHistory(history, true /*remove duplicates*/);

			mainWindow->logConsoleMessage(widget->m_history.asText(
						"Plot widget history: \n", "End plot widget history:\n"));

			// And now only the innermost ranges of the whole history.

			mainWindow->logConsoleMessage(widget->m_history.innermostRangesAsText());

			// The plot widget reacts to the Qt::Key_Delete key by calling a function
			// that emits a set of signals, of which the destroyPlotWidget signal
			// actually indirectly triggers the destruction of the plot widget. This
			// signal is now connected to the main window where a function will call for
			// the destruction of all the target plot widgets, that is all the plot
			// widgets that were created as the result of an integration computation of
			// the plot widget currently being destroyed.

			connect(widget, &AbstractPlotWidget::destroyPlotWidget, mainWindow,
					&MainWindow::destroyAlsoAllTargetPlotWidgets);

			connect(widget, &AbstractPlotWidget::toggleMultiGraphGraphSignal,
					mainWindow, &MainWindow::toggleMultiGraphGraph);

			connect(widget, &MassSpecPlotWidget::newDriftSpectrumByMz,
					mainWindow->mp_driftSpecWnd, &DriftSpecWnd::newDriftSpectrumByMz);

			connect(widget, &MassSpecPlotWidget::newTicChromatogramByMz,
					mainWindow->mp_ticChromWnd, &TicChromWnd::newTicChromatogramByMz);


			// Finally make the mapping between the tic chrom window plot widget and
			// this new mass spec window plot widget.
			AbstractPlotWidget *senderPlotWidget =
				static_cast<AbstractPlotWidget *>(QObject::sender());
			if(senderPlotWidget == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			mainWindow->m_dataPlotWidgetRelationer.newRelation(
					massSpecDataSet, senderPlotWidget, widget,
					nullptr /* open spectra dialog list widget item */);

			// Ensure that there is a connection between this plot widget and this
			// containing window, such that if there is a requirement to replot, the
			// widget gets the signal.
			connect(this, &AbstractMultiPlotWnd::replotWithAxisRange, widget,
					&AbstractPlotWidget::replotWithAxisRange);

			// Make sure we get the signal requiring to redraw the background of the
			// widget, depending on it being focused or not.
			connect(this, &AbstractMultiPlotWnd::redrawPlotBackground, widget,
					&AbstractPlotWidget::redrawPlotBackground);
		}


	void
		MassSpecWnd::newMassSpectrumByDt(MassSpecDataSet *massSpecDataSet,
				IntegrationType integrationType,
				History history, QColor color)
		{
			if(massSpecDataSet == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			m_wishedColor = color;

			// We get a History reference that informs us about the ranges in the
			// sense that the innermost range for integrationType is by necessity the
			// integration range we work on.

			double rangeStart;
			double rangeEnd;

			if(!history.innermostRange(integrationType, &rangeStart, &rangeEnd))
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// qDebug() << __FILE__ << __LINE__
			//<< "rangeStart:" << rangeStart << "rangeEnd:" << rangeEnd;

			MainWindow *mainWindow = static_cast<MainWindow *>(parent());
			QString msgSkeleton =
				QString("Computing mass spectrum between dt %1 and dt %2")
				.arg(rangeStart, 0, 'f', 6)
				.arg(rangeEnd, 0, 'f', 6);

			mainWindow->logConsoleMessage(msgSkeleton);

#if 0
			// No more of use here at this is performed in the calling function.

			History localHistory = history;

			HistoryItem *histItem = new HistoryItem;
			histItem->newIntegrationRange(IntegrationType::DT_TO_MZ, dtStart, dtEnd);

			// Aggreate the new history item to the history we got as a parameter and
			// that we copied into localHistory.

			localHistory.appendHistoryItem(histItem);
#endif

			// Make sure the cancel button is visible.
			mp_statusBarCancelPushButton->setVisible(true);

			double start = omp_get_wtime();

			QVector<double> keyVector;
			QVector<double> valVector;

			if(massSpecDataSet->isStreamed())
			{
				newStreamedMassSpectrum(massSpecDataSet, integrationType, &keyVector,
						&valVector, history, color);
			}
			else // that is, !if(massSpecDataSet->isStreamed())
			{

				// At this point, we already know that we have to perform the work on
				// the basis of data that are in the RAM already.

				// We receive two drift time values that mark a drift time range into
				// which to compute a mass spectrum. In this version of the function we
				// do go to the ground data without filtering by mz values.


				// We are doing an integration DT_TO-MZ, that is that we are selecting
				// drfit times as defined by a drift time range : [dtStart-dtEnd]. This
				// is fine, but where are these data from ? Did these data come from a
				// previous integration or did they come right over the initial RT ?
				// Further, was there a single previous integration or were there
				// multiple ones ? And what integrations were these ? DT-based ? Or
				// what ?

				// Let's see an example : start with the initial RT. Make a first
				// integration to MZ with a RT range. Then, in the MZ spectrum, make an
				// integration to DT with a MZ range. Finally, in the DT spectrum, make
				// an integration to MZ with a DT range. We need to always document what
				// we do and check each time all the previous integrations so as to
				// provide the expected result.

				// DT integration.
				// ===============
				// Already set at the top of this function.

				// MZ integration.
				// ===============
				bool mzIntegration = false;
				double mzStart = -1;
				double mzEnd = -1;

				// If the history is comprised of many successive integrations on the
				// basis of m/z value ranges, then we should make sure we do take into
				// account the more internal m/z value range.
				mzIntegration = history.innermostMzRange(&mzStart, &mzEnd);
				if(mzIntegration)
				{
					mainWindow->logConsoleMessage(
							"m/z integration accounted for: \n" +
							QString("[%1-%2]\n").arg(mzStart).arg(mzEnd));
				}

				// RT integration.
				// ===============
				bool rtIntegration = false;
				double rtStart = -1;
				double rtEnd = -1;

				// If the history is comprised of many successive integrations on the
				// basis of dt value ranges, then we should make sure we do take into
				// account the more internal dt value range.
				rtIntegration = history.innermostRtRange(&rtStart, &rtEnd);
				if(rtIntegration)
				{
					mainWindow->logConsoleMessage(
							"rt integration accounted for: \n" +
							QString("[%1-%2]\n").arg(rtStart).arg(rtEnd));
				}

				// Get handles to the original data in the parent window.
				const msXpSlibmass::MsMultiHash &rtHash = massSpecDataSet->rtHash();

				// Get a list of all the retention times that are stored in the whole
				// file and that are contained in the dt range passed as parameters.

				QList<double> rtList;
				if(rtIntegration)
					rtHash.rangedKeys(rtList, rtStart, rtEnd);
				else
					rtList = rtHash.uniqueKeys();

				qSort(rtList.begin(), rtList.end());

				int rtListSize = rtList.size();

				// qDebug() << __FILE__ << __LINE__
				// << "Number of unique retention time keys: " << rtListSize
				// << "from scan start time "<< rtList.first()
				// << "to scan start time" << rtList.last();

				setupProgressFeedback(msgSkeleton, 1, rtListSize);

				// We are to produce a mass spectrum that is the result of combinations
				// of spectra, so allocate one on the stack (quick calcs).

				msXpSlibmass::MassSpectrum combinedMassSpectrum;

				int iterationCount = 0;

#pragma omp parallel for
				for(int iter = 0; iter < rtListSize; ++iter)
				{

					if(m_isOperationCancelled)
						continue;

					double rt = rtList.at(iter);

					msXpSlibmass::MassSpectrumList rtMassSpectra(rtHash.values(rt));

					for(int jter = 0; jter < rtMassSpectra.size(); ++jter)
					{
						msXpSlibmass::MassSpectrum *spectrum = rtMassSpectra.at(jter);

						// Old version when msXpSlibmass::MassSpectrum did not have m_rt/m_dt member
						// variables.
						// double dt = dtHash.key(spectrum);

						if(spectrum->dt() < rangeStart || spectrum->dt() > rangeEnd)
						{
							// Remove the mass spectrum from the dtMassSpectra because
							// it does
							// not have the proper dt value.
							rtMassSpectra.removeAt(jter);
							--jter;
							continue;
						}
					}

					// At this point, the rtMassSpectra only contain spectra that match
					// both the drift time range and the retention time range. Now,
					// perform the spectra combination by taking into account an
					// eventual
					// MZ-based integration.

					QString msg = msgSkeleton +
						QString("-- rt %1 with %2 spectra")
						.arg(rt, 0, 'f', 6)
						.arg(rtMassSpectra.size());

#pragma omp critical
					{
						if(mzIntegration)
						{
							// We need to use a mz-specific combine function because each
							// MassPeak
							// must be checked for its mz value that must be in between
							// mzStart and
							// mzEnd to be combined.
							combinedMassSpectrum.combine(rtMassSpectra, mzStart, mzEnd);
						}
						else
						{
							combinedMassSpectrum.combine(rtMassSpectra);
						}

						++iterationCount;

						// The label and progress bar widgets were setup by the master
						// thread,
						// so we can only modify them (actually emitting signals) from
						// that
						// same master thread:
						if(omp_get_thread_num() == 0)
							updateProgressFeedback(msg, iterationCount);
					}
				}
				// End of
				// for(int iter = 0; iter < rtListSize; ++iter)

				keyVector = QVector<double>::fromList(combinedMassSpectrum.mzList());
				valVector = QVector<double>::fromList(combinedMassSpectrum.iList());
			}
			// End of
			// else // that is, !if(massSpecDataSet->isStreamed())

			double end = omp_get_wtime();

			QString msg = QString("Computed mass spectrum in %1 s").arg(end - start);

			// Show 100% completion of the work.
			setupProgressFeedback(msg, 0, 100, 100);

			// Hide the cancel button.
			mp_statusBarCancelPushButton->setVisible(false);

			mainWindow->logConsoleMessage(msg);

			// qDebug() << __FILE__ << __LINE__ << msg;

			// In some seconds, hide the progress bar and set the label to an empty
			// text.
			QTimer::singleShot(3000, this, [=]() {
					this->updateProgressFeedback("", 0);
					this->mp_statusBarProgressBar->setVisible(false);
					});

			// Sanity check
			if(keyVector.size() != valVector.size())
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// We may have gotten here because the operation was cancelled. In any case,
			// reset the boolean value.

			m_isOperationCancelled = false;

			// We need to plot that stuff into a new widget.

			// Create a text item to indicate the range.

			QString text = QString("%1\n%2 - drift time range : [%3-%4] ")
				.arg(massSpecDataSet->fileName())
				.arg(m_desc)
				.arg(rangeStart, 0, 'f', 6)
				.arg(rangeEnd, 0, 'f', 6);

			// Allocate the new plot widget in which the data are to be plot.
			MassSpecPlotWidget *widget = static_cast<MassSpecPlotWidget *>(
					addPlotWidget(keyVector, valVector, text, massSpecDataSet));

			// The history that we get as a parameter needs to be appended to the new
			// widget's history so that we have a full account of the history.

			widget->m_history.copyHistory(history, true /*remove duplicates*/);

			mainWindow->logConsoleMessage(widget->m_history.asText(
						"Plot widget history: \n", "End plot widget history:\n"));

			// And now only the innermost ranges of the whole history.

			mainWindow->logConsoleMessage(widget->m_history.innermostRangesAsText());

			// The plot widget reacts to the Qt::Key_Delete key by calling a function
			// that emits a set of signals, of which the destroyPlotWidget signal
			// actually indirectly triggers the destruction of the plot widget. This
			// signal is now connected to the main window where a function will call
			// for the destruction of all the target plot widgets, that is all the
			// plot widgets that were created as the result of an integration
			// computation of the plot widget currently being destroyed.

			connect(widget, &AbstractPlotWidget::destroyPlotWidget, mainWindow,
					&MainWindow::destroyAlsoAllTargetPlotWidgets);

			connect(widget, &AbstractPlotWidget::toggleMultiGraphGraphSignal,
					mainWindow, &MainWindow::toggleMultiGraphGraph);

			connect(widget, &MassSpecPlotWidget::newDriftSpectrumByMz,
					mainWindow->mp_driftSpecWnd, &DriftSpecWnd::newDriftSpectrumByMz);

			connect(widget, &MassSpecPlotWidget::newTicChromatogramByMz,
					mainWindow->mp_ticChromWnd, &TicChromWnd::newTicChromatogramByMz);



			// Finally make the mapping between the drift spec plot widget and
			// this new mass spec plot widget.

			AbstractPlotWidget *senderPlotWidget =
				dynamic_cast<AbstractPlotWidget *>(QObject::sender());
			if(senderPlotWidget == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			mainWindow->m_dataPlotWidgetRelationer.newRelation(
					massSpecDataSet, senderPlotWidget, widget,
					nullptr /* open spectra dialog list widget item */);

			// Ensure that there is a connection between this plot widget and this
			// containing window, such that if there is a requirement to replot, the
			// widget gets the signal.

			connect(this, &AbstractMultiPlotWnd::replotWithAxisRange, widget,
					&AbstractPlotWidget::replotWithAxisRange);

			// Make sure we get the signal requiring to redraw the background of the
			// widget, depending on it being focused or not.

			connect(this, &AbstractMultiPlotWnd::redrawPlotBackground, widget,
					&AbstractPlotWidget::redrawPlotBackground);
		}

	void
		MassSpecWnd::newMassSpectrumByDtAndMz(MassSpecDataSet *massSpecDataSet,
				IntegrationType integrationType,
				History history, QColor color)
		{
			if(massSpecDataSet == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			m_wishedColor = color;

			// We get a History reference that informs us about the ranges in the
			// sense that the innermost range for integrationType is by necessity the
			// integration range we work on. Note this is a particular case because we
			// need to get two pairs of values since we are called by signal emission
			// from the ColorMapPlotWidget, where two dimensions are simulatenously
			// taken into account: the DT and the MZ.

			double dtRangeStart;
			double dtRangeEnd;

			double mzRangeStart;
			double mzRangeEnd;

			if(!history.innermostRange(IntegrationType::DATA_TO_DT_MZ_DT_RANGE,
						&dtRangeStart, &dtRangeEnd))
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			if(!history.innermostRange(IntegrationType::DATA_TO_DT_MZ_MZ_RANGE,
						&mzRangeStart, &mzRangeEnd))
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			MainWindow *mainWindow = static_cast<MainWindow *>(parent());
			QString msgSkeleton =
				QString("Computing mass spectrum between dt %1 and dt %2"
						"and between m/z %3 and m/z %4")
				.arg(dtRangeStart, 0, 'f', 6)
				.arg(dtRangeEnd, 0, 'f', 6)
				.arg(mzRangeStart, 0, 'f', 6)
				.arg(mzRangeEnd, 0, 'f', 6);

			mainWindow->logConsoleMessage(msgSkeleton);

#if 0
			// No more of use here at this is performed in the calling function.

			// Make a local copy of the history that was passed as parameter.
			History localHistory = history;

			// Create a new history item to document this current computation.
			HistoryItem *histItem = new HistoryItem;

			// Because we have integrated using both DT and MZ, we'll have to create
			// two integration ranges for the same history item.

			// First the DT integration range:
			histItem->newIntegrationRange(IntegrationType::DATA_TO_DT_MZ_DT_RANGE,
					dtStart, dtEnd);

			// And now the MZ integration range:
			histItem->newIntegrationRange(IntegrationType::DATA_TO_DT_MZ_MZ_RANGE,
					mzStart, mzEnd);

			// Aggreate the new history item to the history we got as a parameter and
			// that we copied into localHistory.

			localHistory.appendHistoryItem(histItem);
#endif

			// Make sure the cancel button is visible.
			mp_statusBarCancelPushButton->setVisible(true);

			double start = omp_get_wtime();

			QVector<double> keyVector;
			QVector<double> valVector;

			if(massSpecDataSet->isStreamed())
			{
				newStreamedMassSpectrum(massSpecDataSet, integrationType, &keyVector,
						&valVector, history, color);
			}
			else // that is, !if(massSpecDataSet->isStreamed())
			{

				// At this point, we already know that we have to perform the work on
				// the basis of data that are in the RAM already.

				// We receive two drift time values that mark a drift time range into
				// which to compute a mass spectrum. In this version of the function we
				// do go to the ground data without filtering by mz values.


				// We are doing an integration DT_TO-MZ, that is that we are selecting
				// drfit times as defined by a drift time range : [dtStart-dtEnd]. This
				// is fine, but where are these data from ? Did these data come from a
				// previous integration or did they come right over the initial RT ?
				// Further, was there a single previous integration or were there
				// multiple ones ? And what integrations were these ? DT-based ? Or
				// what ?

				// Let's see an example : start with the initial RT. Make a first
				// integration to MZ with a RT range. Then, in the MZ spectrum, make an
				// integration to DT with a MZ range. Finally, in the DT spectrum, make
				// an integration to MZ with a DT range. We need to always document what
				// we do and check each time all the previous integrations so as to
				// provide the expected result.

				// DT integration.
				// ===============
				// Already set at the top of this function.

				// MZ integration.
				// ===============
				// Already set at the top of this function.

				// RT integration.
				// ===============
				bool rtIntegration = false;
				double rtStart = -1;
				double rtEnd = -1;

				// qDebug() << __FILE__ << __LINE__ << "rtIntegration:" << rtIntegration
				//<< "rtStart:" << rtStart << "rtEnd:" << rtEnd;

				// If the history is comprised of many successive integrations on the
				// basis of dt value ranges, then we should make sure we do take into
				// account the more internal dt value range.
				rtIntegration = history.innermostRtRange(&rtStart, &rtEnd);
				if(rtIntegration)
				{
					mainWindow->logConsoleMessage(
							"rt integration accounted for: \n" +
							QString("[%1-%2]\n").arg(rtStart).arg(rtEnd));
				}

				// Get handles to the original data in the parent window.
				const msXpSlibmass::MsMultiHash &rtHash = massSpecDataSet->rtHash();

				// Get a list of all the retention times that are stored in the whole
				// file
				// and that are contained in the dt range passed as parameters.
				QList<double> rtList;
				if(rtIntegration)
					rtHash.rangedKeys(rtList, rtStart, rtEnd);
				else
					rtList = rtHash.uniqueKeys();

				qSort(rtList.begin(), rtList.end());

				int rtListSize = rtList.size();

				// qDebug() << __FILE__ << __LINE__
				// << "Number of unique retention time keys: " << rtListSize
				// << "from scan start time "<< rtList.first()
				// << "to scan start time" << rtList.last();

				setupProgressFeedback(msgSkeleton, 1, rtListSize);

				// We are to produce a mass spectrum that is the result of combinations
				// of
				// spectra, so allocate one on the stack (quick calcs).

				msXpSlibmass::MassSpectrum combinedMassSpectrum;

				int iterationCount = 0;

#pragma omp parallel for
				for(int iter = 0; iter < rtListSize; ++iter)
				{

					if(m_isOperationCancelled)
						continue;

					double rt = rtList.at(iter);

					msXpSlibmass::MassSpectrumList rtMassSpectra(rtHash.values(rt));

					for(int jter = 0; jter < rtMassSpectra.size(); ++jter)
					{
						msXpSlibmass::MassSpectrum *spectrum = rtMassSpectra.at(jter);

						// Old version when msXpSlibmass::MassSpectrum did not have m_rt/m_dt member
						// variables.
						// double dt = dtHash.key(spectrum);

						if(spectrum->dt() < dtRangeStart || spectrum->dt() > dtRangeEnd)
						{
							// Remove the mass spectrum from the dtMassSpectra because
							// it does
							// not have the proper dt value.
							rtMassSpectra.removeAt(jter);
							--jter;
							continue;
						}
					}

					// At this point, the rtMassSpectra only contain spectra that match
					// both the drift time range and the retention time range. Now,
					// perform the spectra combination by taking into account an
					// eventual
					// MZ-based integration.

					QString msg = msgSkeleton +
						QString("-- rt %1 with %2 spectra")
						.arg(rt, 0, 'f', 6)
						.arg(rtMassSpectra.size());

#pragma omp critical
					{
						combinedMassSpectrum.combine(rtMassSpectra, mzRangeStart,
								mzRangeEnd);

						++iterationCount;

						// The label and progress bar widgets were setup by the master
						// thread,
						// so we can only modify them (actually emitting signals) from
						// that
						// same master thread:
						if(omp_get_thread_num() == 0)
							updateProgressFeedback(msg, iterationCount);
					}
				}
				// End of
				// for(int iter = 0; iter < rtListSize; ++iter)

				keyVector = QVector<double>::fromList(combinedMassSpectrum.mzList());
				valVector = QVector<double>::fromList(combinedMassSpectrum.iList());
			}
			// End of
			// else // that is, !if(massSpecDataSet->isStreamed())

			double end = omp_get_wtime();

			QString msg = QString("Computed mass spectrum in %1 s").arg(end - start);

			// Show 100% completion of the work.
			setupProgressFeedback(msg, 0, 100, 100);

			// Hide the cancel button.
			mp_statusBarCancelPushButton->setVisible(false);

			mainWindow->logConsoleMessage(msg);

			// qDebug() << __FILE__ << __LINE__ << msg;

			// In some seconds, hide the progress bar and set the label to an empty
			// text.
			QTimer::singleShot(3000, this, [=]() {
					this->updateProgressFeedback("", 0);
					this->mp_statusBarProgressBar->setVisible(false);
					});

			// Sanity check
			if(keyVector.size() != valVector.size())
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// We may have gotten here because the operation was cancelled. In any case,
			// reset the boolean value.

			m_isOperationCancelled = false;

			// We need to plot that stuff into a new widget.

			// Create a text item to indicate the range.

			QString text = QString("%1\n%2 - drift time range : [%3-%4] ")
				.arg(massSpecDataSet->fileName())
				.arg(m_desc)
				.arg(dtRangeStart, 0, 'f', 6)
				.arg(dtRangeEnd, 0, 'f', 6);

			// Allocate the new plot widget in which the data are to be plot.
			MassSpecPlotWidget *widget = static_cast<MassSpecPlotWidget *>(
					addPlotWidget(keyVector, valVector, text, massSpecDataSet));

			// The history that we get as a parameter needs to be appended to the new
			// widget's history so that we have a full account of the history.

			widget->m_history.copyHistory(history, true /*remove duplicates*/);

			mainWindow->logConsoleMessage(widget->m_history.asText(
						"Plot widget history: \n", "End plot widget history:\n"));

			// And now only the innermost ranges of the whole history.

			mainWindow->logConsoleMessage(widget->m_history.innermostRangesAsText());

			// The plot widget reacts to the Qt::Key_Delete key by calling a function
			// that emits a set of signals, of which the destroyPlotWidget signal
			// actually indirectly triggers the destruction of the plot widget. This
			// signal is now connected to the main window where a function will call
			// for the destruction of all the target plot widgets, that is all the
			// plot widgets that were created as the result of an integration
			// computation of the plot widget currently being destroyed.

			connect(widget, &AbstractPlotWidget::destroyPlotWidget, mainWindow,
					&MainWindow::destroyAlsoAllTargetPlotWidgets);

			connect(widget, &AbstractPlotWidget::toggleMultiGraphGraphSignal,
					mainWindow, &MainWindow::toggleMultiGraphGraph);

			connect(widget, &MassSpecPlotWidget::newDriftSpectrumByMz,
					mainWindow->mp_driftSpecWnd, &DriftSpecWnd::newDriftSpectrumByMz);

			connect(widget, &MassSpecPlotWidget::newTicChromatogramByMz,
					mainWindow->mp_ticChromWnd, &TicChromWnd::newTicChromatogramByMz);


			// Finally make the mapping between the drift spec plot widget and
			// this new mass spec plot widget.
			AbstractPlotWidget *senderPlotWidget =
				dynamic_cast<AbstractPlotWidget *>(QObject::sender());
			if(senderPlotWidget == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			mainWindow->m_dataPlotWidgetRelationer.newRelation(
					massSpecDataSet, senderPlotWidget, widget,
					nullptr /* open spectra dialog list widget item */);

			// Ensure that there is a connection between this plot widget and this
			// containing window, such that if there is a requirement to replot, the
			// widget gets the signal.
			connect(this, &AbstractMultiPlotWnd::replotWithAxisRange, widget,
					&AbstractPlotWidget::replotWithAxisRange);

			// Make sure we get the signal requiring to redraw the background of the
			// widget, depending on it being focused or not.
			connect(this, &AbstractMultiPlotWnd::redrawPlotBackground, widget,
					&AbstractPlotWidget::redrawPlotBackground);
		}


	// We are not going to produce any graph, we are integrating the data to a
	// single TIC intensity value.
	void 
		MassSpecWnd::ticIntensity(MassSpecDataSet *massSpecDataSet,
				double *intensity, IntegrationType integrationType, History history)
		{
			if(massSpecDataSet == Q_NULLPTR)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
			if(intensity == Q_NULLPTR)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// We get a History reference that informs us about the ranges in the
			// sense that the innermost range for integrationType is by necessity the
			// integration range we work on.

			double rangeStart;
			double rangeEnd;

			if(!history.innermostRange(integrationType, &rangeStart, &rangeEnd))
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			MainWindow *mainWindow = static_cast<MainWindow *>(parent());
			QString msgSkeleton =
				QString("Computing TIC intensity between m/z %1 and m/z %2")
				.arg(rangeStart, 0, 'f', 6)
				.arg(rangeEnd, 0, 'f', 6);
			mainWindow->logConsoleMessage(msgSkeleton);

			// Make sure the cancel button is visible.
			mp_statusBarCancelPushButton->setVisible(true);

			double start = omp_get_wtime();

			if(massSpecDataSet->isStreamed())
			{
				streamedTicIntensity(massSpecDataSet, 
						intensity, integrationType, history);
			}
			else // that is, !if(massSpecDataSet->isStreamed())
			{
				// RT integration.
				// ===============
				bool rtIntegration = false;
				double rtStart = -1;
				double rtEnd = -1;

				// If the history is comprised of many successive integrations on the
				// basis of rt value ranges, then we should make sure we do take into
				// account the more internal rt value range.
				rtIntegration = history.innermostRtRange(&rtStart, &rtEnd);
				if(rtIntegration)
				{
					mainWindow->logConsoleMessage(
							"rt integration accounted for: \n" +
							QString("[%1-%2]\n").arg(rtStart).arg(rtEnd));
				}

				// MZ integration.
				// ===============
				// Already set at the top of this function.

				// DT integration.
				// ===============
				bool dtIntegration = false;
				double dtStart = -1;
				double dtEnd = -1;

				// If the history is comprised of many successive integrations on the
				// basis of dt value ranges, then we should make sure we do take into
				// account the more internal dt value range.
				dtIntegration = history.innermostDtRange(&dtStart, &dtEnd);
				if(dtIntegration)
				{
					mainWindow->logConsoleMessage(
							"dt integration accounted for: \n" +
							QString("[%1-%2]\n").arg(dtStart).arg(dtEnd));
				}

				// Get handles to the original data in the calling entity.
				const msXpSlibmass::MsMultiHash &rtHash = massSpecDataSet->rtHash();

				// One way to start the computation is simply to define which retention
				// times should go in the Tic intensity calculation.

				QList<double> rtList;
				if(rtIntegration)
					rtHash.rangedKeys(rtList, rtStart, rtEnd);
				else
					rtList = rtHash.uniqueKeys();

				qSort(rtList.begin(), rtList.end());

				int rtListSize = rtList.size();

				// qDebug() << __FILE__ << __LINE__
				// << "Number of unique retention time keys: " << rtListSize
				// << "from scan start time "<< rtList.first()
				// << "to scan start time" << rtList.last();

				setupProgressFeedback(msgSkeleton, 1, rtListSize);

				// For each retention time, we'll want to
				// 1. get the corresponding mass spectra
				// 2. only retain the spectra that match the drift time (if applicable)
				// 3. compute a TIC value for the retained spectra that is summed 
				// to all the preceding ones, to finally yield a total Tic intensity
				// value (that actually is simply an integration).

				double localIntensity = 0.0;
				int iterationCount = 0;

#pragma omp parallel for
				for(int iter = 0; iter < rtListSize; ++iter)
				{

					if(m_isOperationCancelled)
						continue;

					double rt = rtList.at(iter);

					msXpSlibmass::MassSpectrumList rtMassSpectra(rtHash.values(rt));

					if(dtIntegration)
					{

						// We have to remove from all the rtMassSpectra, all the spectra
						// that are not in the drift time range.

						for(int jter = 0; jter < rtMassSpectra.size(); ++jter)
						{
							msXpSlibmass::MassSpectrum *spectrum = rtMassSpectra.at(jter);

							// Old version when msXpSlibmass::MassSpectrum did not have m_rt/m_dt
							// member
							// variables.
							// double dt = dtHash.key(spectrum);

							if(spectrum->dt() < dtStart || spectrum->dt() > dtEnd)
							{
								// Remove the mass spectrum from the rtMassSpectra
								// because it does
								// not have the proper dt value.
								rtMassSpectra.removeAt(jter);
								--jter;
								continue;
							}
						}
					}

					// At this point, the rtMassSpectra only contain spectra that match
					// both the retention time range and the drift time range.

					for(int kter = 0; kter < rtMassSpectra.size(); ++kter)
					{
						double newTic = 0.0;
						newTic = rtMassSpectra.at(kter)->tic(rangeStart, rangeEnd);

#pragma omp critical
						localIntensity += newTic;
					}
					// End of for(int kter = 0; kter < massSpectra.size(); ++kter)

					QString msg = msgSkeleton +
						QString("-- rt %1 with %2 spectra")
						.arg(rt, 0, 'f', 6)
						.arg(rtMassSpectra.size());

#pragma omp critical
					{
						++iterationCount;

						// The label and progress bar widgets were setup by the master
						// thread, so we can only modify them (actually emitting signals)
						// from that same master thread:

						if(omp_get_thread_num() == 0)
							updateProgressFeedback(msg, iterationCount);
					}
				}
				// End of
				// #pragma omp parallel for
				// for(int iter = 0; iter < rtList.size(); ++iter)

				// Finally, update the value passed to the function:
				*intensity = localIntensity;
			}
			// End of
			// else // that is, !if(massSpecDataSet->isStreamed())

			double end = omp_get_wtime();

			double duration = end - start;

			QString msg = QString("Took %1 s to compute TIC intensity: %2\n")
				.arg(duration)
				.arg(*intensity, 0, 'g', 3);

			// Show 100% completion of the work.
			setupProgressFeedback(msg, 0, 100, 100);

			// Hide the cancel button.
			mp_statusBarCancelPushButton->setVisible(false);

			mainWindow->logConsoleMessage(msg);

			// qDebug() << __FILE__ << __LINE__ << msg;

			// Pretty quickly hide the progress bar and set the label to an empty
			// text so that the user can see the TIC I value on the statusbar.

			// The idea is that if the computation is very quick, then the message
			// should vanish almost immediately. Otherwise let it there longer to show
			// to the user that the computation is computing-intensive.

			// For every second of integration, show the message 20 ms. But do not
			// exceed 3000 ms.
			int timeOut = duration * 20; 
			if(timeOut > 3000)
				timeOut = 3000;

			QTimer::singleShot(timeOut, this, [=]() {
					this->updateProgressFeedback("", 0);
					this->mp_statusBarProgressBar->setVisible(false);
					});

			// We may have gotten here because the operation was cancelled. In any
			// case, reset the boolean value.

			m_isOperationCancelled = false;
		}


	void
		MassSpecWnd::newStreamedMassSpectrum(MassSpecDataSet *massSpecDataSet,
				IntegrationType integrationType,
				QVector<double> *keyVector,
				QVector<double> *valVector,
				History history, QColor color)
		{

			// Depending on the format of the data file from which the data were
			// loaded in stream mode, it will be possible or not to perform the
			// required integration.

			if(massSpecDataSet->fileFormat() ==
					MassSpecDataFileFormat::MASS_SPEC_DATA_FILE_FORMAT_SQLITE3)
			{

				qDebug() << __FILE__ << __LINE__
					<< "the filename at this stage is:" << massSpecDataSet->fileName();

				MassSpecDataFileLoaderSqlite3 loader(massSpecDataSet->fileName());

				// Make sure we connect the feedback signal/slot
				connect(&loader,
						&MassSpecDataFileLoaderSqlite3::setupProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::setupProgressFeedback);

				connect(&loader,
						&MassSpecDataFileLoaderSqlite3::updateProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::updateProgressFeedback);

				connect(this, &AbstractMultiPlotWnd::cancelOperationSignal, &loader,
						&MassSpecDataFileLoaderSqlite3::cancelOperation);

				int res = loader.streamedIntegration(
						history, IntegrationType::ANY_TO_MZ, keyVector, valVector);

				if(res == -1)
					qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
			}
			else if(massSpecDataSet->fileFormat() ==
					MassSpecDataFileFormat::MASS_SPEC_DATA_FILE_FORMAT_MZML)
			{
				MassSpecDataFileLoaderPwiz loader(massSpecDataSet->fileName());

				// Make sure we connect the feedback signal/slot
				connect(&loader,
						&MassSpecDataFileLoaderPwiz::setupProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::setupProgressFeedback);

				connect(&loader,
						&MassSpecDataFileLoaderPwiz::updateProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::updateProgressFeedback);

				connect(this, &AbstractMultiPlotWnd::cancelOperationSignal,
						&loader, &MassSpecDataFileLoaderPwiz::cancelOperation);

				int res = loader.streamedIntegration(
						history, IntegrationType::ANY_TO_MZ, keyVector, valVector);

				if(res == -1)
					qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
			}
			else if(massSpecDataSet->fileFormat() ==
					MassSpecDataFileFormat::MASS_SPEC_DATA_FILE_FORMAT_DX)
			{
				MassSpecDataFileLoaderDx loader(massSpecDataSet->fileName());

				// Make sure we connect the feedback signal/slot
				connect(&loader,
						&MassSpecDataFileLoaderDx::setupProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::setupProgressFeedback);

				connect(&loader,
						&MassSpecDataFileLoaderDx::updateProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::updateProgressFeedback);


				connect(this, &AbstractMultiPlotWnd::cancelOperationSignal,
						&loader, &MassSpecDataFileLoaderDx::cancelOperation);

				int res = loader.streamedIntegration(
						history, keyVector, valVector, Q_NULLPTR /* intensity */);

				if(res == -1)
					qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
			}
			else if(massSpecDataSet->fileFormat() ==
					MassSpecDataFileFormat::MASS_SPEC_DATA_FILE_FORMAT_MS1)
			{
				MassSpecDataFileLoaderMs1 loader(massSpecDataSet->fileName());

				// Make sure we connect the feedback signal/slot
				connect(&loader,
						&MassSpecDataFileLoaderMs1::setupProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::setupProgressFeedback);

				connect(&loader,
						&MassSpecDataFileLoaderMs1::updateProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::updateProgressFeedback);

				connect(this, &AbstractMultiPlotWnd::cancelOperationSignal,
						&loader, &MassSpecDataFileLoaderMs1::cancelOperation);

				int res = loader.streamedIntegration(
						history, keyVector, valVector, Q_NULLPTR /* intensity */);

				if(res == -1)
					qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
			}
			else
			{
				QMessageBox msgBox;
				msgBox.setText(
						"The operation cannot be carried over because the "
						"data were loaded in stream and not from a usable file.");

				msgBox.setStandardButtons(QMessageBox::Ok);
				msgBox.exec();

				// Hide the cancel button.
				mp_statusBarCancelPushButton->setVisible(false);

				return;
			}
		}


	void
		MassSpecWnd::streamedTicIntensity(MassSpecDataSet *massSpecDataSet,
				double *intensity,IntegrationType integrationType, History history)
		{
			if(intensity == Q_NULLPTR)
				qFatal("Fatal error at %s@%d. Program aborted.",
						__FILE__, __LINE__);

			// Depending on the format of the data file from which the data were
			// loaded in stream mode, it will be possible or not to perform the
			// required integration.

			if(massSpecDataSet->fileFormat() ==
					MassSpecDataFileFormat::MASS_SPEC_DATA_FILE_FORMAT_SQLITE3)
			{
				MassSpecDataFileLoaderSqlite3 loader(massSpecDataSet->fileName());

				// Make sure we connect the feedback signal/slot
				connect(&loader,
						&MassSpecDataFileLoaderSqlite3::setupProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::setupProgressFeedback);

				connect(&loader,
						&MassSpecDataFileLoaderSqlite3::updateProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::updateProgressFeedback);

				connect(this, &AbstractMultiPlotWnd::cancelOperationSignal, &loader,
						&MassSpecDataFileLoaderSqlite3::cancelOperation);

				int res = loader.streamedTicIntensity(history, intensity);

				if(res == -1)
					qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
			}
			else if(massSpecDataSet->fileFormat() ==
					MassSpecDataFileFormat::MASS_SPEC_DATA_FILE_FORMAT_MZML)
			{
				MassSpecDataFileLoaderPwiz loader(massSpecDataSet->fileName());

				// Make sure we connect the feedback signal/slot
				connect(&loader,
						&MassSpecDataFileLoaderPwiz::setupProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::setupProgressFeedback);

				connect(&loader,
						&MassSpecDataFileLoaderPwiz::updateProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::updateProgressFeedback);

				connect(this, &AbstractMultiPlotWnd::cancelOperationSignal,
						&loader, &MassSpecDataFileLoaderPwiz::cancelOperation);

				int res = loader.streamedTicIntensity(history, intensity);

				if(res == -1)
					qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
			}
			else if(massSpecDataSet->fileFormat() ==
					MassSpecDataFileFormat::MASS_SPEC_DATA_FILE_FORMAT_DX)
			{
				MassSpecDataFileLoaderDx loader(massSpecDataSet->fileName());

				// Make sure we connect the feedback signal/slot
				connect(&loader,
						&MassSpecDataFileLoaderDx::setupProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::setupProgressFeedback);

				connect(&loader,
						&MassSpecDataFileLoaderDx::updateProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::updateProgressFeedback);


				connect(this, &AbstractMultiPlotWnd::cancelOperationSignal,
						&loader, &MassSpecDataFileLoaderDx::cancelOperation);

				int res = loader.streamedIntegration(history, 
						Q_NULLPTR /* keyVector */, Q_NULLPTR /* valVector */, intensity);

				if(res == -1)
					qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
			}
			else if(massSpecDataSet->fileFormat() ==
					MassSpecDataFileFormat::MASS_SPEC_DATA_FILE_FORMAT_MS1)
			{
				MassSpecDataFileLoaderMs1 loader(massSpecDataSet->fileName());

				// Make sure we connect the feedback signal/slot
				connect(&loader,
						&MassSpecDataFileLoaderMs1::setupProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::setupProgressFeedback);

				connect(&loader,
						&MassSpecDataFileLoaderMs1::updateProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::updateProgressFeedback);

				connect(this, &AbstractMultiPlotWnd::cancelOperationSignal,
						&loader, &MassSpecDataFileLoaderMs1::cancelOperation);

				int res = loader.streamedIntegration(
						history, Q_NULLPTR /* keyVector */, Q_NULLPTR /* valVector */, intensity);

				if(res == -1)
					qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
			}
			else
			{
				QMessageBox msgBox;
				msgBox.setText(
						"The operation cannot be carried over because the "
						"data were loaded in stream and not from a database file.");

				msgBox.setStandardButtons(QMessageBox::Ok);
				msgBox.exec();

				// Hide the cancel button.
				mp_statusBarCancelPushButton->setVisible(false);

				return;
			}
		}


	void
		MassSpecWnd::recordAnalysisStanza(const QString &stanza, const QColor &color)
		{
			// The record might go to a file or to the console or both.

			if((mp_analysisPreferences->m_recordTarget &
						RecordTarget::RECORD_TO_FILE) == RecordTarget::RECORD_TO_FILE)
			{
				if(mp_analysisFile != nullptr)
				{

					// Write the stanza, that was crafted by the calling plot widget to
					// the file.

					if(!mp_analysisFile->open(QIODevice::Append))
					{
						statusBar()->showMessage(
								QString("Could not record the step because "
									"the file could not be opened."),
								4000);
					}
					else
					{
						mp_analysisFile->write(stanza.toLatin1());
						// Force writing because we may want to have tail -f work fine
						// on the
						// file, and see modifications live to change fiels in a text
						// editor.
						mp_analysisFile->flush();
						mp_analysisFile->close();
					}
				}
				else
				{
					qDebug() << __FILE__ << __LINE__
						<< "The mp_analysisFile pointer is nullptr.";

					statusBar()->showMessage(
							QString("Could not record the analysis step to file. "
								"Please define a file to write the data to."),
							4000);
				}
			}

			// Also, if recording to the console is asked for, then do that also.
			if((mp_analysisPreferences->m_recordTarget &
						RecordTarget::RECORD_TO_CONSOLE) == RecordTarget::RECORD_TO_CONSOLE)
			{
				MainWindow *mainWindow = static_cast<MainWindow *>(parent());
				mainWindow->logConsoleMessage(stanza, color);
			}

			// Also, if recording to the clipboard is asked for, then do that also.
			if((mp_analysisPreferences->m_recordTarget &
						RecordTarget::RECORD_TO_CLIPBOARD) == RecordTarget::RECORD_TO_CLIPBOARD)
			{
				QClipboard *clipboard = QApplication::clipboard();
				clipboard->setText(clipboard->text() + stanza, QClipboard::Clipboard);
			}

			return;
		}

} // namespace msXpSmineXpert
