/*  BEGIN software license
 *
 *  msXpertSuite - mass spectrometry software suite
 *  -----------------------------------------------
 *  Copyright(C) 2009, 2017 Filippo Rusconi
 *
 *  http://www.msxpertsuite.org
 *
 *  This file is part of the msXpertSuite project.
 *
 *  The msXpertSuite project is the successor of the massXpert project. This
 *  project now includes various independent modules:
 *  
 *  - massXpert, model polymer chemistries and simulate mass spectrometric data;
 *  - mineXpert, a powerful TIC chromatogram/mass spectrum viewer/miner;
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * END software license
*/



/////////////////////// Qt includes
#include <QObject>
#include <QDebug>
#include <QDialog>
#include <QWidget>
#include <QColorDialog>
#include <QPen>


/////////////////////// Local includes
#include "OpenSpectraDlg.hpp"
#include "MassSpecDataSet.hpp"
#include "AbstractPlotWidget.hpp"
#include "MainWindow.hpp"
#include "DataPlotWidgetRelation.hpp"


namespace msXpSmineXpert
{

	OpenSpectraDlg::OpenSpectraDlg(QWidget *parent) : QDialog{parent}
	{
		if(parent == nullptr)
			qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

		m_ui.setupUi(this);

		if(!initialize())
			qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
	}


	OpenSpectraDlg::~OpenSpectraDlg()
	{
	}


	void
		OpenSpectraDlg::closeEvent(QCloseEvent *event)
		{
			writeSettings();
			hide();
		}


	void
		OpenSpectraDlg::writeSettings()
		{
			QSettings settings;
			settings.setValue("OpenSpectraDlg_geometry", saveGeometry());
		}


	void
		OpenSpectraDlg::readSettings()
		{
			QSettings settings;
			restoreGeometry(settings.value("OpenSpectraDlg_geometry").toByteArray());
		}


	bool
		OpenSpectraDlg::initialize()
		{
			setWindowTitle("mineXpert - Loaded mass spectrum files");

			// Populate the action combo box.

			m_ui.actionComboBox->addItems(m_actionList);

			readSettings();

			// Make the connections

			connect(
					m_ui.actionComboBox,
					static_cast<void (QComboBox::*)(const QString &)>(&QComboBox::activated),
					this, &OpenSpectraDlg::actionComboBoxActivated);

			return true;
		}


	QListWidgetItem *
		OpenSpectraDlg::addOpenSpectrum(MassSpecDataSet *dataSet, QColor color)
		{
			if(dataSet == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// Create an icon that is simply a rectangle filled with color.
			QPixmap rectangle(15, 15);
			rectangle.fill(color);
			QIcon icon(rectangle);

			QListWidgetItem *item = new QListWidgetItem(icon, dataSet->fileName());

			// The font color must match the color that was used to display the
			// corresponding graph.

			QBrush brush(Qt::SolidPattern);
			brush.setColor(color);

			item->setForeground(brush);

			// Append the new items to the current list.
			int row = m_ui.openSpectraListWidget->count();
			m_ui.openSpectraListWidget->insertItem(row, item);

			// Get the TIC chrom wnd's plot widget pointer so that we can store it in
			// the map along with the pointer to the list widget item. This mapping
			// will allow to remove the list widget item when the plot widget is
			// destroyed.
			AbstractPlotWidget *widget =
				dynamic_cast<AbstractPlotWidget *>(dataSet->parent());
			if(widget == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			m_listItemTicPlotMap.insert(item, widget);

			return item;
		}


	void
		OpenSpectraDlg::removeListWidgetItem(AbstractPlotWidget *widget)
		{
			// Get the widget item corresponding to the plot widget:
			QListWidgetItem *item = m_listItemTicPlotMap.key(widget);
			m_listItemTicPlotMap.remove(item);

			// Deleting the list widget item will remove it automatically from the
			// widget list.
			delete item;
		}


	void
		OpenSpectraDlg::actionComboBoxActivated(const QString &text)
		{
			MainWindow *mainWindow = static_cast<MainWindow *>(parent());

			if(text == "Hide" || text == "Show" || text == "Toggle visibility")
			{
				// When the user asks for plot widgets to be hidden, what she asks for
				// is that the matching multigraph graph be hidden. But we never hide
				// the graph in the singlegraph plot widgets of the various windows
				// (mass, drift or tic window).
				QList<QListWidgetItem *> selectedItemList =
					m_ui.openSpectraListWidget->selectedItems();

				if(selectedItemList.size() == 0)
					return;


				qDebug() << __FILE__ << __LINE__ << "selected item count"
					<< selectedItemList.size();

				for(int iter = 0; iter < selectedItemList.size(); ++iter)
				{
					QListWidgetItem *curItem = selectedItemList.at(iter);

					// This is the corresponding TIC chrom window's plot widget. This
					// is the root widget because that is the widget that is created
					// first right after having loaded data from the file. That is, all
					// the other widgets are targets of this origin widget (for a
					// single file, of course, more than one file might have been
					// opened, in which case we have more than on TicChromPlotWidget
					// listed in this dialog list widget and we are in the loop
					// iterating in these sequentially).

					AbstractPlotWidget *widget = m_listItemTicPlotMap.value(curItem);

					QList<AbstractPlotWidget *> targetPlotWidgetList;
					mainWindow->m_dataPlotWidgetRelationer.targetsRecursive(
							widget, &targetPlotWidgetList);

					for(int jter = 0; jter < targetPlotWidgetList.size(); ++jter)
					{
						AbstractPlotWidget *widget = targetPlotWidgetList.at(jter);

						qDebug() << __FILE__ << __LINE__ << "here";

						if(text == "Hide")
							widget->hideMultiGraphGraph();
						else if(text == "Show")
							widget->showMultiGraphGraph();
						else if(text == "Toggle visibility")
							widget->toggleMultiGraphGraph();
					}
				}
			}
			else if(text == "Hide all" || text == "Show all")
			{
				// All the items must be hidden or shown.
				QList<AbstractPlotWidget *> allPlotWidgetList;
				int itemCount = m_ui.openSpectraListWidget->count();

				for(int iter = 0; iter < itemCount; ++iter)
				{
					QListWidgetItem *curItem = m_ui.openSpectraListWidget->item(iter);

					// This is the corresponding TIC chrom window's plot widget.
					AbstractPlotWidget *widget = m_listItemTicPlotMap.value(curItem);

					QList<AbstractPlotWidget *> targetPlotWidgetList;
					mainWindow->m_dataPlotWidgetRelationer.targetsRecursive(
							widget, &targetPlotWidgetList);

					for(int jter = 0; jter < targetPlotWidgetList.size(); ++jter)
					{
						AbstractPlotWidget *widget = targetPlotWidgetList.at(jter);

						if(text == "Hide all")
							widget->hideMultiGraphGraph();
						else
							widget->showMultiGraphGraph();
					}
				}
			}
			else if(text == "Hide all but selected" || text == "Show all but selected")
			{
				int itemCount = m_ui.openSpectraListWidget->count();

				for(int iter = 0; iter < itemCount; ++iter)
				{
					QListWidgetItem *curItem = m_ui.openSpectraListWidget->item(iter);

					// This is the corresponding TIC chrom window's plot widget.
					AbstractPlotWidget *widget = m_listItemTicPlotMap.value(curItem);

					QList<AbstractPlotWidget *> targetPlotWidgetList;
					mainWindow->m_dataPlotWidgetRelationer.targetsRecursive(
							widget, &targetPlotWidgetList);

					for(int jter = 0; jter < targetPlotWidgetList.size(); ++jter)
					{
						AbstractPlotWidget *widget = targetPlotWidgetList.at(jter);

						if(text == "Hide all but selected")
						{
							if(curItem->isSelected())
								widget->showMultiGraphGraph();
							else
								widget->hideMultiGraphGraph();
						}
						else
						{
							if(curItem->isSelected())
								widget->hideMultiGraphGraph();
							else
								widget->showMultiGraphGraph();
						}
					}
				}
			}
			else if(text == "Change color")
			{

				// This option can only be carried over if there is a single item
				// selected.

				QList<QListWidgetItem *> selectedItemList =
					m_ui.openSpectraListWidget->selectedItems();

				if(selectedItemList.size() == 0 || selectedItemList.size() > 1)
					return;

				// This is the TIC chrom plot widget.

				AbstractPlotWidget *plotWidget = firstSelectedPlotWidget();

				// Store the current color for later recycling.
				QColor currentColor = plotWidget->graph()->pen().color();

				// Let the user select the color.

				QColor newColor = QColorDialog::getColor(Qt::white, nullptr, 
						QString("Please select a new color for the plot"));

				if(!newColor.isValid())
					return;

				// We need the main window to get access to the color handling features.

				MainWindow *mainWindow = static_cast<MainWindow *>(parent());

				QColor *useColor = nullptr;

				// Let's see if the color is already used.

				if(mainWindow->isPlottingColorUsed(newColor))
				{
					QMessageBox::StandardButton ret = 
						QMessageBox::warning
						(this, "mineXpert - Open spectra",
						 "This color is already used. Go on ?",
						 QMessageBox::Yes | QMessageBox::No | QMessageBox::Cancel);

					if(ret == QMessageBox::Cancel)
						return ;
					else if (ret == QMessageBox::No)
						return;
				}
				else
				{
					useColor = new QColor(newColor);
					mainWindow->m_usedColorList.append(useColor);
				}

				const DataPlotWidgetRelationer &relationer = mainWindow->m_dataPlotWidgetRelationer;
				QList<DataPlotWidgetRelation *> relationList;
				AbstractPlotWidget *recoloringPlotWidget = nullptr;

				relationer.getRelationsByOriginRecursive(plotWidget, &relationList);

				qDebug() << __FILE__ << __LINE__
					<< "The number of relations is:" << relationList.size();

				for(int iter = 0; iter < relationList.size(); ++iter)
				{
					DataPlotWidgetRelation *relation = relationList.at(iter);

					// A DataPlotWidgetRelation relates plot widgets together.

					recoloringPlotWidget = relation->mp_originPlotWidget;
					if(recoloringPlotWidget != nullptr)
						recoloringPlotWidget->changeColor(newColor);

					recoloringPlotWidget = relation->mp_targetPlotWidget;
					if(recoloringPlotWidget != nullptr)
						recoloringPlotWidget->changeColor(newColor);
				}

				// And now finish by recoloring the color vignette of the corresponding
				// item in the QListWidget and the matching file name text.
				QListWidgetItem *item = selectedItemList.first();

				qDebug() << __FILE__ << __LINE__
					<< "here";

				// Create an icon that is simply a rectangle filled with color.
				QPixmap rectangle(15, 15);
				rectangle.fill(newColor);
				QIcon icon(rectangle);

				item->setIcon(icon);

				// The font color must match the color that was used to display the
				// corresponding graph.

				QBrush brush(Qt::SolidPattern);
				brush.setColor(newColor);
				item->setForeground(brush);

				// Finally, recycle the previous color
				mainWindow->recyclePlottingColor(currentColor);
			}
			else if(text == "Close")
			{
				// The currently selected items must be closed.

				QList<QListWidgetItem *> selectedItemList =
					m_ui.openSpectraListWidget->selectedItems();

				if(selectedItemList.size() == 0)
					return;

				for(int iter = 0; iter < selectedItemList.size(); ++iter)
				{
					QListWidgetItem *curItem = selectedItemList.at(iter);

					// This is the corresponding TIC chrom window's plot widget.
					AbstractPlotWidget *widget = m_listItemTicPlotMap.value(curItem);

					// The call below will initiate a cascade of events, first removing
					// the matching multigraph graph, then triggering the removal of
					// this plot widget. This removal will first trigger the removal of
					// any target plot widget related to this origin plot widget (the
					// origin plot widget is
					// the plot widget in which the user started an integration
					// operation that then produces a new plot widget with the results
					// of the integration, like DT->MS, for example. The MS plot widget
					// is a target plot widget and the DT plot widget is an origin plot
					// widget. Each time a root plot widget is removed, automatically
					// the matching widget list
					// item is deleted, so we do not need to handle anything here, else
					// than asking the widget to initiate that cascade of events.

					widget->shouldDestroyPlotWidget();
				}
			}
			else if(text == "Close all but selected")
			{
				int itemCount = m_ui.openSpectraListWidget->count();

				// We cannot remove the widget inside the for loop below because when we
				// destroy a widget, the QListWidgetItem is destroyed.
				QList<AbstractPlotWidget *> plotWidgetList;

				for(int iter = 0; iter < itemCount; ++iter)
				{
					QListWidgetItem *curItem = m_ui.openSpectraListWidget->item(iter);

					if(!curItem->isSelected())
					{
						AbstractPlotWidget *widget =
							m_listItemTicPlotMap.value(curItem);

						plotWidgetList.append(widget);
					}
				}

				// Now that we have a list of plot widgets to remove we can iterate in
				// the list and destroy the widgets sequentially.
				for(int jter = 0; jter < plotWidgetList.size(); ++jter)
					plotWidgetList.at(jter)->shouldDestroyPlotWidget();
			}
			else if(text == "Close all")
			{
				while(m_ui.openSpectraListWidget->count())
				{
					QListWidgetItem *curItem = m_ui.openSpectraListWidget->item(0);

					// This is the corresponding TIC chrom window's plot widget.
					AbstractPlotWidget *widget = m_listItemTicPlotMap.value(curItem);

					widget->shouldDestroyPlotWidget();
				}
			}
		}


	bool
		OpenSpectraDlg::isPlotWidgetSelected(AbstractPlotWidget *widget)
		{
			if(widget == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			QListWidgetItem *item = m_listItemTicPlotMap.key(widget);
			if(item->isSelected())
				return true;

			return false;
		}


	int
		OpenSpectraDlg::itemCount()
		{
			return m_ui.openSpectraListWidget->count();
		}


	int
		OpenSpectraDlg::selectedItemCount()
		{
			return m_ui.openSpectraListWidget->selectedItems().size();
		}


	int
		OpenSpectraDlg::firstSelectedPlotWidgetIndex(AbstractPlotWidget *widget)
		{
			// If there are multiple selected items, return the one with the least
			// index and set widget to the corresponding first selected item.

			QList<QListWidgetItem *> itemList =
				m_ui.openSpectraListWidget->selectedItems();
			QListWidgetItem *item = nullptr;

			if(itemList.size() == 0)
			{
				widget = nullptr;
				return -1;
			}

			item = itemList.at(0);

			if(widget != nullptr)
				widget = m_listItemTicPlotMap.value(item);

			return m_ui.openSpectraListWidget->row(item);
		}


	AbstractPlotWidget *
		OpenSpectraDlg::firstSelectedPlotWidget(int *index)
		{
			// If there are multiple selected items, return the one with the least
			// index.

			QList<QListWidgetItem *> itemList =
				m_ui.openSpectraListWidget->selectedItems();

			AbstractPlotWidget *widget = nullptr;
			QListWidgetItem *   item   = nullptr;

			if(itemList.size() == 0)
			{
				if(index != nullptr)
					*index = -1;

				return widget;
			}

			item = itemList.at(0);

			if(index != nullptr)
				*index = m_ui.openSpectraListWidget->row(item);

			return m_listItemTicPlotMap.value(item);
		}


	AbstractPlotWidget *
		OpenSpectraDlg::plotWidgetAt(int index)
		{
			if(index < 0)
				return nullptr;

			if(index >= m_ui.openSpectraListWidget->count())
				return nullptr;

			// Get the plot widget pointer on the basis of the list widet item at
			// index.
			QListWidgetItem *item = m_ui.openSpectraListWidget->item(index);

			AbstractPlotWidget *widget = m_listItemTicPlotMap.value(item);

			return widget;
		}


	AbstractPlotWidget *
		OpenSpectraDlg::selectPlotWidget(int index)
		{
			if(index < 0)
				return nullptr;

			if(index >= m_ui.openSpectraListWidget->count())
				return nullptr;

			// Get the plot widget pointer on the basis of the list widet item at
			// index.
			QListWidgetItem *item = m_ui.openSpectraListWidget->item(index);

			item->setSelected(true);

			AbstractPlotWidget *widget = m_listItemTicPlotMap.value(item);

			return widget;
		}


	int
		OpenSpectraDlg::selectPlotWidget(AbstractPlotWidget *widget)
		{
			if(widget == nullptr)
				return -1;

			QListWidgetItem *item = m_listItemTicPlotMap.key(widget);

			item->setSelected(true);

			return m_ui.openSpectraListWidget->row(item);
		}


	AbstractPlotWidget *
		OpenSpectraDlg::togglePlotWidget(int index)
		{
			if(index < 0)
				return nullptr;

			if(index >= m_ui.openSpectraListWidget->count())
				return nullptr;

			// Get the plot widget pointer on the basis of the list widet item at
			// index.
			QListWidgetItem *item = m_ui.openSpectraListWidget->item(index);

			item->setSelected(!item->isSelected());

			AbstractPlotWidget *widget = m_listItemTicPlotMap.value(item);

			return widget;
		}


	int
		OpenSpectraDlg::togglePlotWidget(AbstractPlotWidget *widget)
		{
			if(widget == nullptr)
				return -1;

			QListWidgetItem *item = m_listItemTicPlotMap.key(widget);
			item->setSelected(!item->isSelected());

			return m_ui.openSpectraListWidget->row(item);
		}


	int
		OpenSpectraDlg::selectAllPlotWidgets()
		{
			int iter      = 0;
			int itemCount = m_ui.openSpectraListWidget->count();
			for(; iter < itemCount; ++iter)
				m_ui.openSpectraListWidget->item(iter)->setSelected(true);

			return itemCount;
		}


	int
		OpenSpectraDlg::deselectAllPlotWidgets()
		{
			int iter      = 0;
			int itemCount = m_ui.openSpectraListWidget->count();

			for(; iter < itemCount; ++iter)
				m_ui.openSpectraListWidget->item(iter)->setSelected(false);

			return itemCount;
		}


	AbstractPlotWidget *
		OpenSpectraDlg::deselectAllPlotWidgetsButOne(int index)
		{
			AbstractPlotWidget *widget = nullptr;
			QListWidgetItem *   item   = nullptr;

			for(int iter = 0; iter < m_ui.openSpectraListWidget->count(); ++iter)
			{
				item = m_ui.openSpectraListWidget->item(iter);

				if(iter == index)
				{
					widget = m_listItemTicPlotMap.value(item);

					item->setSelected(true);
				}
				else
					item->setSelected(false);
			}

			return widget;
		}


	int
		OpenSpectraDlg::deselectAllPlotWidgetsButOne(AbstractPlotWidget *widget)
		{
			if(widget == nullptr)
				return -1;

			AbstractPlotWidget *curWidget = nullptr;
			int                 itemIndex = -1;

			for(int iter = 0; iter < m_ui.openSpectraListWidget->count(); ++iter)
			{
				QListWidgetItem *item = m_ui.openSpectraListWidget->item(iter);
				curWidget             = m_listItemTicPlotMap.value(item);

				if(curWidget == widget)
				{
					item->setSelected(true);
					itemIndex = iter;
				}
				else
					item->setSelected(false);
			}

			return itemIndex;
		}


	void
		OpenSpectraDlg::keyReleaseEvent(QKeyEvent *event)
		{
			//qDebug() << __FILE__ << __LINE__ << "OpenSpectraDlg::keyReleaseEvent"
			//<< "event key:" << event->key();

			if(event->key() == Qt::Key_T)
				actionComboBoxActivated("Toggle");
			else if(event->key() == Qt::Key_H)
				actionComboBoxActivated("Hide");
			else if(event->key() == Qt::Key_S)
				actionComboBoxActivated("Show");
		}


} // namespace msXpSmineXpert
