/*  BEGIN software license
 *
 *  msXpertSuite - mass spectrometry software suite
 *  -----------------------------------------------
 *  Copyright(C) 2009, 2017 Filippo Rusconi
 *
 *  http://www.msxpertsuite.org
 *
 *  This file is part of the msXpertSuite project.
 *
 *  The msXpertSuite project is the successor of the massXpert project. This
 *  project now includes various independent modules:
 *  
 *  - massXpert, model polymer chemistries and simulate mass spectrometric data;
 *  - mineXpert, a powerful TIC chromatogram/mass spectrum viewer/miner;
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * END software license
 */



/////////////////////// Qt includes
#include <QPushButton>


/////////////////////// Local includes
#include "globals.hpp"
#include "TicChromWnd.hpp"
#include "Application.hpp"
#include "MainWindow.hpp"
#include "libmass/MsMultiHash.hpp"
#include "MassSpecDataFileLoaderPwiz.hpp"
#include "MassSpecDataFileLoaderDx.hpp"
#include "MassSpecDataFileLoaderMs1.hpp"


#include <omp.h>

namespace msXpSmineXpert
{

	int TicChromWnd::widgetCounter = 0;

	TicChromWnd::TicChromWnd(QWidget *parent)
		: AbstractMultiPlotWnd{parent, "mineXpert" /*m_name*/,
			"TIC-chromatogram" /*m_desc*/}
	{
		initialize();
	}

	TicChromWnd::~TicChromWnd()
	{
	}

	void
		TicChromWnd::initialize()
		{
			initializePlotRegion();

			show();
		}



	void
		TicChromWnd::initializePlotRegion()
		{
			// There are two regions in the window. The upper part of the mp_splitter
			// is gonna receive the multi-graph plotwidget. The lower part of the
			// splitter will receive all the individual plot widgets.

			mp_multiGraphPlotWidget =
				new TicChromPlotWidget(this, this, m_name, m_desc);
			mp_multiGraphPlotWidget->setObjectName("multiGraphPlot");


			// We will need later to know that this plot widget is actually for
			// displaying many different graphs, contrary to the normal situation.
			mp_multiGraphPlotWidget->m_isMultiGraph = true;

			// We need to remove the single graph that was created upon creation of
			// the QCustomPlot widget because we'll later add graphs, and we want to
			// have a proper correlation between the index of the graphs and their
			// pointer value. If we did not remove that graph "by default", then there
			// would be a gap of 1 between the index of the first added mass graph and
			// its expected value.
			mp_multiGraphPlotWidget->clearGraphs();

			// Ensure that there is a connection between  this plot widget and this
			// containing window, such that if there is a requirement to replot, the
			// widget gets the signal.
			connect(this, &TicChromWnd::replotWithAxisRange, mp_multiGraphPlotWidget,
					&TicChromPlotWidget::replotWithAxisRange);

			// This window is responsible for the writing of the analysis stanzas to
			// the file. We need to catch the signals.
			connect(mp_multiGraphPlotWidget, &TicChromPlotWidget::recordAnalysisStanza,
					this, &TicChromWnd::recordAnalysisStanza);

			// This window is responsible for displaying the key/value pairs of the
			// graphs onto which the mouse is moved. We need to catch the signal.
			connect(mp_multiGraphPlotWidget, &AbstractPlotWidget::updateStatusBarSignal, this,
					&AbstractMultiPlotWnd::updateStatusBar);

			// We want to create a scroll area in which the multigraph plot widget
			// will be sitting.
			QScrollArea *scrollArea = new QScrollArea();
			scrollArea->setObjectName(QStringLiteral("scrollArea"));
			scrollArea->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOn);
			scrollArea->setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOn);
			// We accept that the plots are resized, but we set a minimum size when we
			// create them.
			scrollArea->setWidgetResizable(true);
			scrollArea->setAlignment(Qt::AlignCenter);
			scrollArea->setWidget(mp_multiGraphPlotWidget);
			// And now prepend to the mp_splitter that new scroll area. Remember that
			// the mp_splitter had already a widget set to it, that is, the scroll
			// area that is now located below the mp_multiGraphPlotWidget, that is
			// responsible for the hosting of all the single-graph plot widgets.
			mp_splitter->insertWidget(0, scrollArea);
		}


	TicChromPlotWidget *
		TicChromWnd::addPlotWidget(QVector<double> keyVector,
				QVector<double> valVector, const QString &desc,
				MassSpecDataSet *massSpecDataSet)
		{
			MainWindow *mainWindow = static_cast<MainWindow *>(parent());

			// Beware that massSpecDataSet might be nullptr.
			QString fileName;
			if(massSpecDataSet != nullptr)
				fileName = massSpecDataSet->fileName();

			TicChromPlotWidget *widget = new TicChromPlotWidget(
					this, this, m_name, desc, fileName, massSpecDataSet);
			widget->setObjectName("monoGraphPlot");


			///////////////// BEGIN Script availability of the new widget //////////////////

			// Craft a name that differentiates this widget from the previous and next
			// ones. We use the static counter number to affix to the string
			// "ticChrom".

			QString propName =
				QString("ticChromPlot%1").arg(TicChromWnd::widgetCounter);
			QString alias = "lastTicChromPlot";

			QString explanation = 
				QString("Make available a TIC chromatogram plot under name \"%1\" "
						"with alias \"%2\".\n").arg(propName).arg(alias);

			// Make the new plot widget globally available under the crafted property
			// name and alias. Also, put out the explanation text. Note that by
			// providing 'this', we make the new object appear in the scripting objet
			// tree widget as a child item of this.

			QScriptValue scriptValue;
			mainWindow->mp_scriptingWnd->publishQObject(widget, 
					this /* parent object */, 
					propName, alias, explanation, QString() /* help string */,
					scriptValue,
					QScriptEngine::QtOwnership, 
					QScriptEngine::QObjectWrapOptions());

#if 0
			// Old way of doing, we now handle the alias in one go above

			//Now make the new plot available under the name "lastTicChromPlot"
			//for the user to catch it with own variable, because that object name
			//will be overwritten later.

			mainWindow->mp_scriptingWnd->publishQObject(widget, this  parent object ,
					"lastTicChromPlot", QString()  alias ,
					"The last (in time) TIC chromatogram plot created. Use this handle at "
					"creation time to store the object under a name of your choice",
					"Make available the last TIC chromatogram plot under name \"lastTicChromPlot\"",
					scriptValue);
#endif

			++TicChromWnd::widgetCounter;

			///////////////// END Script availability of the new widget //////////////////



			widget->setMinimumSize(400, 150);
			mp_scrollAreaVBoxLayout->addWidget(widget);

			widget->graph()->setData(keyVector, valVector);

			// Let's see if one color was set as the desired color, otherwise use
			// black.
			QPen currentPen = widget->graph()->pen();

			if(!m_wishedColor.isValid())
				currentPen.setColor(Qt::black);
			else
				currentPen.setColor(m_wishedColor);

			widget->graph()->setPen(currentPen);

			widget->rescaleAxes(true /*only enlarge*/);
			widget->replot();

			// Store in the history the new ranges of both x and y axes.
			widget->updateAxisRangeHistory();

			// At this point, make sure that the same data are represented on the
			// multi-graph plot widget.

			QCPGraph *newGraph = mp_multiGraphPlotWidget->addGraph();
			newGraph->setData(keyVector, valVector);
			newGraph->setPen(currentPen);
			mp_multiGraphPlotWidget->rescaleAxes(true /*only enlarge*/);
			mp_multiGraphPlotWidget->replot();

			// Maintain a hash connection between any given plot widget and the
			// corresponding graph in the multigraph widget.
			m_plotWidgetGraphMap.insert(widget, newGraph);

			// Finally append the new widget to the list of widgets that belong to
			// this window.
			m_plotWidgetList.append(widget);

			// Signal/slot connections.

			connect(widget, &TicChromPlotWidget::recordAnalysisStanza, this,
					&TicChromWnd::recordAnalysisStanza);

			connect(widget, &AbstractPlotWidget::updateStatusBarSignal, this,
					&AbstractMultiPlotWnd::updateStatusBar);

			return widget;
		}

	AbstractPlotWidget *
		TicChromWnd::initialTic(MassSpecDataSet *massSpecDataSet)
		{
			if(massSpecDataSet == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// There are two kinds of situation:
			//
			// 1. The data are loaded in a streamed manner, that is the whole set of
			// mass spectra from the file is not stored in RAM. While loading the
			// data, the loader has dealt with TIC data and crafted a TIC chromatogram
			// in the massSpecDataSet.
			//
			// 2. The data are fully loaded in RAM and thus we use them to calculate
			// the TIC chromatogram now.

			// A total ion current chromatogram is created by iteratinng in each
			// retention
			// time, getting all the spectra for that retention time and then summing-up
			// for
			// these spectra the intensities of all the ions (that is all the peaks). We
			// have a QMultiHash that relates the various retention times to their mass
			// spectra. Just iterate in the retention times and do  the mass spec
			// combine.

			MainWindow *mainWindow = static_cast<MainWindow *>(parent());

			QVector<double> keyVector;
			QVector<double> valVector;

			if(!massSpecDataSet->isStreamed())
			{
				// Get a const handle on the ground data, that is the data stored after
				// loading from the file.

				const msXpSlibmass::MsMultiHash &rtHash = massSpecDataSet->rtHash();

				QList<double> rtList = rtHash.uniqueKeys();
				qSort(rtList.begin(), rtList.end());

				int rtListSize = rtList.size();

				// qDebug() << __FILE__ << __LINE__
				//<< "Number of unique retention time keys: " << rtListSize;

				setupProgressFeedback("Initial TIC chrom. calculation", 1, rtListSize);

				double start = omp_get_wtime();

				// We need a map, because the parallelization below will iterate in
				// rtList in disorder and we won't be able to rely on the iter value to
				// have rt/tic colinearity as we used to do in previous non-parallel
				// versions.

				QMap<double, double> rtTicMap;

				// We cannot rely on the parallel code to iterated in increasing iter
				// values below, so we need to have a specific counter that will be
				// incremented at each iteration.
				int iterationCount = 0;

#pragma omp parallel
				{
#pragma omp for
					for(int iter = 0; iter < rtListSize; ++iter)
					{
						double rt = rtList.at(iter);

						// This mass spectrum list only holds pointers to MassSpectrum
						// instances allocated on the heap, but it does not own them.

						msXpSlibmass::MassSpectrumList rtMassSpectra(rtHash.values(rt)); 

						// qDebug() << __FILE__ << __LINE__
						// << "For retention time: " << rt
						// << "number of mass spectra: " << rtMassSpectra.size();

						// We now have to iterate in all these spectra, seek for the
						// MassPeak
						// intensity of all the spectra and sum that up to yield a single
						// total
						// ion current value for that retention time.

						double totalTic = msXpSlibmass::MassSpectrumList::tic(rtMassSpectra);

						QString msg = QString("Retention time: %1 -> %2 spectra")
							.arg(rt, 0, 'f', 6)
							.arg(rtMassSpectra.size());

						// qDebug() << __FILE__ << __LINE__
						//<< msg;


#pragma omp critical
						{
							rtTicMap.insert(rt, totalTic);

							++iterationCount;

							// The label and progress bar widgets were setup by the master
							// thread,
							// so we can only modify them (actually emitting signals) from
							// that
							// same master thread:
							if(omp_get_thread_num() == 0)
								updateProgressFeedback(msg, iterationCount);
						}
					}
					// End of
					// for(int iter = 0; iter < rtListSize; ++iter)
				}

				double end = omp_get_wtime();

				QString msg =
					QString("Computed TIC chromatogram in %1 s.").arg(end - start);

				// Because only the master thread can update the feedback data, the
				// loop might end while we have not reached an iterationCount
				// corresponding to 100%. This might make the user uncomfortable. So we
				// simulate the 100% achievement:

				updateProgressFeedback(msg, rtListSize);

				QTimer::singleShot(3000, this, [=]() {
						this->updateProgressFeedback("", 0);
						this->mp_statusBarProgressBar->setVisible(false);
						});

				mainWindow->logConsoleMessage(msg);

				// qDebug() << __FILE__ << __LINE__ << msg;

				keyVector = QVector<double>::fromList(rtTicMap.keys());
				valVector = QVector<double>::fromList(rtTicMap.values());
			}
			else
			{

				// If the file was opened in the streaming mode, then, by definition,
				// the TIC chromatogram will have been calculated upon load and is
				// available in the massSpecDataSet. All we have to do is to plot these
				// data.

				keyVector =
					QVector<double>::fromList(massSpecDataSet->ticChromMap().keys());
				valVector =
					QVector<double>::fromList(massSpecDataSet->ticChromMap().values());

				// qDebug() << __FILE__ << __LINE__
				//<< "Streamed mode vectors: keys: " << keyVector.size()
				//<< "elements:" << keyVector
				//<< "and values: " << valVector.size() << "elements"
				//<< valVector;
			}

			// qDebug() << __FILE__ << __LINE__ << "keyVector:" << keyVector
			//<< "valVector:" << valVector;

			bool needsFakingData = false;

			if(keyVector.size() == 0)
			{
				QMessageBox msgBox;
				msgBox.setText(
						"The program could not load data from the file. Please,"
						" ensure that the file is correct.");

				msgBox.setStandardButtons(QMessageBox::Ok);
				msgBox.exec();

				return Q_NULLPTR;
			}

			if(keyVector.size() == 1)
			{
				// It might be possible that the key/val vectors have only one value if
				// the user
				// loaded an asc|txt|xy file where the data are only in the form of two
				// columns of numbers. In this case  the single point present in the
				// vectors is at key=0 and val the summed value of all the numbers in
				// the second column of the data. We should show meaninful data.

				// qDebug() << __FILE__ << __LINE__ << "Need to fake some data";

				fakeData(&keyVector, &valVector);
				needsFakingData = true;
			}

			// At this point what we should do is transfer the data to the plotting
			// widget creation routine. Note that we are creating the very first plot
			// widget with data just loaded from a mass spec file. That plot widget
			// will be later connected to the destruction of the mass spec data set
			// when the plot widget itself is deleted (see true param in newRelation
			// call below).

			AbstractPlotWidget *widget = addPlotWidget(
					keyVector, valVector, massSpecDataSet->fileName(), massSpecDataSet);

			// If we needed to fake data to have something visible in the graph, then
			// make a zoom out so that the user is comfortable selecting the graph to
			// make the first integration.
			if(needsFakingData)
			{
				widget->xAxis->setRange(-1, 2);
			}

			// For this specific widget, that is created for storing the initial mass
			// spec data set as read from file, we set it as parent of the mass spec
			// data set, so that when the widget is destroyed, the mass spec data set
			// also.
			massSpecDataSet->setParent(widget);

			// Let mainWindow call the data plot widget relationer because it will
			// have the open spectra dialog widget list item pointer available.
			// mainWindow->m_dataPlotWidgetRelationer.newRelation(massSpecDataSet,

			// At this point, immediately craft  a new HistoryItem instance to append
			// to the widget's History.

			HistoryItem *histItem = new HistoryItem;
			histItem->newIntegrationRange(IntegrationType::DATA_TO_RT,
					keyVector.first(), keyVector.last());
			widget->addHistoryItem(histItem);

			qDebug() << __FILE__ << __LINE__
				<< "After initialTic was computed, history:"
				<< widget->m_history.asText();

			// Log this recent history to the console.
			mainWindow->logConsoleMessage(histItem->asText());
			mainWindow->logConsoleMessage(widget->m_history.innermostRangesAsText());

			// The plot widget reacts to the Qt::Key_Delete key by calling a function
			// that emits a set of signals, of which the destroyPlotWidget signal
			// actually indirectly triggers the destruction of the plot widget. This
			// signal is now connected to the main window where a function will call
			// for the destruction of all the target plot widgets, that is all the
			// plot widgets that were created as the result of an integration
			// computation of the plot widget currently being destroyed.

			connect(widget, &AbstractPlotWidget::destroyPlotWidget, mainWindow,
					&MainWindow::destroyAlsoAllTargetPlotWidgets);

			connect(widget, &AbstractPlotWidget::toggleMultiGraphGraphSignal,
					mainWindow, &MainWindow::toggleMultiGraphGraph);

			// Ensure that there is a connexion between this new widget and the
			// MassSpecWnd instance that will receive the signal requiring to make a
			// mass spectrum on the basis of an integration selection in the TIC
			// chromatogram plot.

			connect(static_cast<TicChromPlotWidget *>(widget),
					&TicChromPlotWidget::newMassSpectrumByRt,
					mainWindow->mp_massSpecWnd, &MassSpecWnd::newMassSpectrumByRt);

			connect(static_cast<TicChromPlotWidget *>(widget),
					&TicChromPlotWidget::newDriftSpectrumByRt,
					mainWindow->mp_driftSpecWnd, &DriftSpecWnd::newDriftSpectrumByRt);

			// Ensure that there is a connection between  this plot widget and this
			// containing window, such that if there is a requirement to replot, the
			// widget gets the signal.
			connect(this, &TicChromWnd::replotWithAxisRange, widget,
					&TicChromPlotWidget::replotWithAxisRange);

			// Make sure we get the signal requiring to redraw the background of the
			// widget, depending on it being focused or not.
			connect(this, &AbstractMultiPlotWnd::redrawPlotBackground, widget,
					&AbstractPlotWidget::redrawPlotBackground);

			return widget;
		}


	void
		TicChromWnd::fakeData(QVector<double> *keyVector, QVector<double> *valVector)
		{
			if(keyVector == nullptr || valVector == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
			if(keyVector->size() > 1 || valVector->size() > 1)
			{

				// qDebug() << __FILE__ << __LINE__ << "The vectors already contain at "
				//"least two points. Doing no fake "
				//"data.";
				return;
			}

			double firstY = valVector->first();

			// The single point in the vectors is at x=0 and y= sum of all the numbers
			// of the second column. We want to have a second point at x=1 and y same
			// value. This is because if there is a single point in the TIC chrom
			// graph, then the user will think that there has been an error. 

			keyVector->append(1.0);
			valVector->append(firstY);
		}


	void
		TicChromWnd::setAnalysisPreferences(AnalysisPreferences *preferences)
		{
			if(preferences == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			mp_analysisPreferences = preferences;
		}


	void
		TicChromWnd::setAnalysisFile(QFile *analysisFile)
		{
			if(analysisFile == nullptr)
				qFatal(
						"Fatal error at %s@%d. The file pointer cannot be 0. Program aborted.",
						__FILE__, __LINE__);

			mp_analysisFile = analysisFile;
		}


	void
		TicChromWnd::newTicChromatogramByMz(MassSpecDataSet *massSpecDataSet,
				IntegrationType  integrationType,
				History history, QColor color)
		{
			if(massSpecDataSet == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			m_wishedColor = color;

			// We get a History reference that informs us about the ranges in the
			// sense that the innermost range for integrationType is by necessity the
			// integration range we work on.

			double rangeStart;
			double rangeEnd;

			if(!history.innermostRange(integrationType, &rangeStart, &rangeEnd))
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			MainWindow *mainWindow = static_cast<MainWindow *>(parent());
			QString     msgSkeleton =
				QString("Computing TIC chromatogram between m/z %1 and m/z %2")
				.arg(rangeStart, 0, 'f', 6)
				.arg(rangeEnd, 0, 'f', 6);
			mainWindow->logConsoleMessage(msgSkeleton);

#if 0
			// No more of use here at this is performed in the calling function.

			History      localHistory = history;
			HistoryItem *histItem     = new HistoryItem;
			histItem->newIntegrationRange(IntegrationType::MZ_TO_RT, mzStart, mzEnd);
			histItem->setDateTime();

			// Aggreate the new history item to the history we got as a parameter and
			// that we copied into localHistory.

			localHistory.appendHistoryItem(histItem);
#endif

			// Make sure the cancel button is visible.
			mp_statusBarCancelPushButton->setVisible(true);

			double start = omp_get_wtime();

			QVector<double> keyVector;
			QVector<double> valVector;

			if(massSpecDataSet->isStreamed())
			{
				newStreamedTicChromatogram(massSpecDataSet, integrationType, &keyVector,
						&valVector, history, color);
			}
			else // that is, !if(massSpecDataSet->isStreamed())
			{
				// At this point, we already know that we have to perform the work on
				// the basis
				// of data that are in the RAM already.

				// We are doing an integration DT_TO-RT, that is that we are selecting
				// drift times as defined by a drift time range : [dtStart-dtEnd].
				// This is fine, but where are these data from ? Did these data come
				// from
				// a previous integration or did they come right over the initial RT ?
				// Further, was there a single previous integration or were there
				// multiple
				// ones ?  And what integrations were these ? MZ-based ? Or what ?

				// Let's see an example : start with the initial RT. Make a first
				// integration to MZ with a RT range. Then, in the MZ spectrum, make an
				// integration to DT with a MZ range. Finally, in the DT spectrum, make
				// an
				// integration to MZ with a DT range. We need to always document what we
				// do and check each time all the previous integrations so as to provide
				// the expected result.

				// RT integration.
				// ===============
				bool   rtIntegration = false;
				double rtStart       = -1;
				double rtEnd         = -1;

				rtIntegration = history.innermostRtRange(&rtStart, &rtEnd);
				if(rtIntegration)
				{
					mainWindow->logConsoleMessage(
							"rt integration accounted for: \n" +
							QString("[%1-%2]\n").arg(rtStart).arg(rtEnd));
				}

				// MZ integration.
				// ===============
				// Already set at the top of this function.

				// DT integration.
				// ===============
				bool   dtIntegration = false;
				double dtStart       = -1;
				double dtEnd         = -1;

				// If the history is comprised of many successive integrations on the
				// basis of dt value ranges, then we should make sure we do take into
				// account the more internal dt value range.
				dtIntegration = history.innermostDtRange(&dtStart, &dtEnd);
				if(dtIntegration)
				{
					mainWindow->logConsoleMessage(
							"dt integration accounted for: \n" +
							QString("[%1-%2]\n").arg(dtStart).arg(dtEnd));
				}

				// Get handles to the original data in the calling entity.
				const msXpSlibmass::MsMultiHash &rtHash = massSpecDataSet->rtHash();

				// Because we ultimately will plot a total ion current value for each
				// retention time (that is we will plot a TIC chromatogram), then we
				// must
				// craft the final list of rt values.
				QList<double> rtList;
				if(rtIntegration)
					rtHash.rangedKeys(rtList, rtStart, rtEnd);
				else
					rtList = rtHash.uniqueKeys();

				qSort(rtList.begin(), rtList.end());

				int rtListSize = rtList.size();

				// qDebug() << __FILE__ << __LINE__
				// << "Number of unique retention time keys: " << rtListSize
				// << "from scan start time "<< rtList.first()
				// << "to scan start time" << rtList.last();

				setupProgressFeedback(msgSkeleton, 1, rtListSize);

				// For each retention time, we'll want to
				// 1. get the corresponding mass spectra
				// 2. only retain the spectra that match the drift time (params to this
				// function)
				// 3. store the rt (retention time) value because this is going to be
				// the
				// x axis key
				// 3. compute a TIC value for the retained spectra that will be the y
				// axis value.
				QMap<double, double> rtTicMap;

				int iterationCount = 0;

#pragma omp parallel for
				for(int iter = 0; iter < rtListSize; ++iter)
				{

					if(m_isOperationCancelled)
						continue;

					double rt = rtList.at(iter);

					msXpSlibmass::MassSpectrumList rtMassSpectra(rtHash.values(rt));

					// We have to remove from all the rtMassSpectra, all the spectra
					// that
					// are not in the drift time range.

					if(dtIntegration)
					{
						// We have to remove from all the rtMassSpectra, all the spectra
						// that
						// are not in the drift time range.

						for(int jter = 0; jter < rtMassSpectra.size(); ++jter)
						{
							msXpSlibmass::MassSpectrum *spectrum = rtMassSpectra.at(jter);

							// Old version when msXpSlibmass::MassSpectrum did  not have m_rt/m_dt
							// member
							// variables.
							// double dt = dtHash.key(spectrum);

							if(spectrum->dt() < dtStart || spectrum->dt() > dtEnd)
							{
								// Remove the mass spectrum from the rtMassSpectra
								// because it does
								// not have the proper dt value.
								rtMassSpectra.removeAt(jter);
								--jter;
								continue;
							}
						}
					}

					// At this point, the rtMassSpectra only contain spectra that match
					// both the retention time range and the drift time range.

					for(int kter = 0; kter < rtMassSpectra.size(); ++kter)
					{
						double newTic = 0.0;
						newTic = rtMassSpectra.at(kter)->tic(rangeStart, rangeEnd);

						// Had another spectrum by the same rt already been iterated in?
						double oldTic = rtTicMap.value(rt, qSNaN());

						if(!qIsNaN(oldTic))
						{

							// A rt by the same value alreay existed, update the tic
							// value and insert it, that will erase the old value,
							// because the map is NOT a multimap.

							newTic += oldTic;
						}

#pragma omp critical
						rtTicMap.insert(rt, newTic);
					}
					// End of for(int kter = 0; kter < massSpectra.size(); ++kter)

					QString msg = msgSkeleton +
						QString("-- rt %1 with %2 spectra")
						.arg(rt, 0, 'f', 6)
						.arg(rtMassSpectra.size());

#pragma omp critical
					{
						++iterationCount;

						// The label and progress bar widgets were setup by the master
						// thread,
						// so we can only modify them (actually emitting signals) from
						// that
						// same master thread:
						if(omp_get_thread_num() == 0)
							updateProgressFeedback(msg, iterationCount);
					}
				}
				// End of for(int iter = 0; iter < rtList.size(); ++iter)

				// The two commands below work because QMap is ordered along the key and
				// returns the keys in ascending order and the values in ascending order
				// *of the keys*.
				keyVector = QVector<double>::fromList(rtTicMap.keys());
				valVector = QVector<double>::fromList(rtTicMap.values());
			}
			// End of
			// else // that is, !if(massSpecDataSet->isStreamed())

			double end = omp_get_wtime();

			QString msg = QString("Computed TIC chromatogram in %1 s").arg(end - start);

			// Show 100% completion of the work.
			setupProgressFeedback(msg, 0, 100, 100);

			// Hide the cancel button.
			mp_statusBarCancelPushButton->setVisible(false);

			mainWindow->logConsoleMessage(msg);

			// qDebug() << __FILE__ << __LINE__ << msg;

			// In some seconds, hide the progress bar and set the label to an empty
			// text.
			QTimer::singleShot(3000, this, [=]() {
					this->updateProgressFeedback("", 0);
					this->mp_statusBarProgressBar->setVisible(false);
					});

			// Sanity check

			if(keyVector.size() != valVector.size())
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// We may have gotten here because the operation was cancelled. In any
			// case, reset the boolean value.

			m_isOperationCancelled = false;

			// We need to plot that stuff into a new widget.

			// Create a text item to indicate the range.

			QString text = QString("%1\n%2 - m/z range : [%3-%4] ")
				.arg(massSpecDataSet->fileName())
				.arg("XIC chromatogram")
				.arg(rangeStart, 0, 'f', 6)
				.arg(rangeEnd, 0, 'f', 6);

			// Allocate the new plot widget in which the data are to be plot.

			TicChromPlotWidget *widget = static_cast<TicChromPlotWidget *>(
					addPlotWidget(keyVector, valVector, text, massSpecDataSet));

			// The history that we get as a parameter needs to be appended to the new
			// widget's history so that we have a full account of the history.

			widget->m_history.copyHistory(history, true /*remove duplicates*/);

			mainWindow->logConsoleMessage(widget->m_history.asText(
						"Plot widget history: \n", "End plot widget history:\n"));

			// And now only the innermost ranges of the whole history.

			mainWindow->logConsoleMessage(widget->m_history.innermostRangesAsText());

			// The plot widget reacts to the Qt::Key_Delete key by calling a function
			// that emits a set of signals, of which the destroyPlotWidget signal
			// actually indirectly triggers the destruction of the plot widget. This
			// signal is now connected to the main window where a function will call for
			// the destruction of all the target plot widgets, that is all the plot
			// widgets that were created as the result of an integration computation of
			// the plot widget currently being destroyed.

			connect(widget, &AbstractPlotWidget::destroyPlotWidget, mainWindow,
					&MainWindow::destroyAlsoAllTargetPlotWidgets);

			connect(widget, &AbstractPlotWidget::toggleMultiGraphGraphSignal,
					mainWindow, &MainWindow::toggleMultiGraphGraph);

			connect(widget, &TicChromPlotWidget::newMassSpectrumByRt,
					mainWindow->mp_massSpecWnd, &MassSpecWnd::newMassSpectrumByRt);

			connect(widget, &TicChromPlotWidget::newDriftSpectrumByRt,
					mainWindow->mp_driftSpecWnd, &DriftSpecWnd::newDriftSpectrumByRt);

			// Finally make the mapping between the tic chrom window plot widget and
			// this new mass spec window plot widget.
			AbstractPlotWidget *senderPlotWidget =
				dynamic_cast<AbstractPlotWidget *>(QObject::sender());
			if(senderPlotWidget == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			mainWindow->m_dataPlotWidgetRelationer.newRelation(
					massSpecDataSet, senderPlotWidget, widget,
					nullptr /* open spectra dialog list widget item */);

			// Ensure that there is a connection between  this plot widget and this
			// containing window, such that if there is a requirement to replot, the
			// widget gets the signal.
			connect(this, &AbstractMultiPlotWnd::replotWithAxisRange, widget,
					&AbstractPlotWidget::replotWithAxisRange);

			// Make sure we get the signal requiring to redraw the background of the
			// widget, depending on it being focused or not.
			connect(this, &AbstractMultiPlotWnd::redrawPlotBackground, widget,
					&AbstractPlotWidget::redrawPlotBackground);
		}


	void
		TicChromWnd::newTicChromatogramByDt(MassSpecDataSet *massSpecDataSet,
				IntegrationType  integrationType,
				History history, QColor color)
		{
			if(massSpecDataSet == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			m_wishedColor = color;

			// We get a History reference that informs us about the ranges in the
			// sense that the innermost range for integrationType is by necessity the
			// integration range we work on.

			double rangeStart;
			double rangeEnd;

			if(!history.innermostRange(integrationType, &rangeStart, &rangeEnd))
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// qDebug() << __FILE__ << __LINE__
			//<< "rangeStart:" << rangeStart << "rangeEnd:" << rangeEnd;

			MainWindow *mainWindow = static_cast<MainWindow *>(parent());
			QString     msgSkeleton =
				QString("Computing TIC chromatogram between dt %1 and dt %2")
				.arg(rangeStart, 0, 'f', 6)
				.arg(rangeEnd, 0, 'f', 6);
			mainWindow->logConsoleMessage(msgSkeleton);

#if 0
			// No more of use here at this is performed in the calling function.

			History      localHistory = history;
			HistoryItem *histItem     = new HistoryItem;
			histItem->newIntegrationRange(IntegrationType::DT_TO_RT, dtStart, dtEnd);
			histItem->setDateTime();

			// Aggreate the new history item to the history we got as a parameter and
			// that we copied into localHistory.

			localHistory.appendHistoryItem(histItem);
#endif

			// Make sure the cancel button is visible.
			mp_statusBarCancelPushButton->setVisible(true);

			double start = omp_get_wtime();

			QVector<double> keyVector;
			QVector<double> valVector;

			if(massSpecDataSet->isStreamed())
			{
				newStreamedTicChromatogram(massSpecDataSet, integrationType, &keyVector,
						&valVector, history, color);
			}
			else // that is, !if(massSpecDataSet->isStreamed())
			{
				// At this point, we already know that we have to perform the work on
				// the basis
				// of data that are in the RAM already.

				// We receive two drift time values that describe the drift times of the
				// drift spectrum window graph and we need to compute the TIC
				// chromatogram
				// of all the ions that have gone through that drift time range.


				// We are doing an integration DT_TO-RT, that is that we are selecting
				// drift times as defined by a drift time range : [dtStart-dtEnd].
				// This is fine, but where are these data from ? Did these data come
				// from
				// a previous integration or did they come right over the initial RT ?
				// Further, was there a single previous integration or were there
				// multiple
				// ones ?  And what integrations were these ? MZ-based ? Or what ?

				// Let's see an example : start with the initial RT. Make a first
				// integration to MZ with a RT range. Then, in the MZ spectrum, make an
				// integration to DT with a MZ range. Finally, in the DT spectrum, make
				// an
				// integration to MZ with a DT range. We need to always document what we
				// do and check each time all the previous integrations so as to provide
				// the expected result.

				// RT integration.
				// ===============
				bool   rtIntegration = false;
				double rtStart       = -1;
				double rtEnd         = -1;

				rtIntegration = history.innermostRtRange(&rtStart, &rtEnd);
				if(rtIntegration)
				{
					mainWindow->logConsoleMessage(
							"rt integration accounted for: \n" +
							QString("[%1-%2]\n").arg(rtStart).arg(rtEnd));
				}

				// MZ integration.
				// ===============
				bool   mzIntegration = false;
				double mzStart       = -1;
				double mzEnd         = -1;

				// If the history is comprised of many successive integrations on the
				// basis of m/z value ranges, then we should make sure we do take into
				// account the more internal m/z value range.
				mzIntegration = history.innermostMzRange(&mzStart, &mzEnd);
				if(mzIntegration)
				{
					mainWindow->logConsoleMessage(
							"m/z integration accounted for: \n" +
							QString("[%1-%2]\n").arg(mzStart).arg(mzEnd));
				}

				// DT integration.
				// ===============
				// Already set at the top of this function.

				// Get handles to the original data in the calling entity.
				const msXpSlibmass::MsMultiHash &rtHash = massSpecDataSet->rtHash();

				// Because we ultimately will plot a total ion current value for each
				// retention time (that is we will plot a TIC chromatogram), then we
				// must
				// craft the final list of rt values.

				QList<double> rtList;
				if(rtIntegration)
					rtHash.rangedKeys(rtList, rtStart, rtEnd);
				else
					rtList = rtHash.uniqueKeys();

				qSort(rtList.begin(), rtList.end());

				int rtListSize = rtList.size();

				// qDebug() << __FILE__ << __LINE__
				// << "Number of unique retention time keys: " << rtListSize
				// << "from scan start time "<< rtList.first()
				// << "to scan start time" << rtList.last();

				setupProgressFeedback(msgSkeleton, 1, rtListSize);

				// For each retention time, we'll want to
				// 1. get the corresponding mass spectra
				// 2. only retain the spectra that match the drift time (params to this
				// function)
				// 3. store the rt (retention time) value because this is going to be
				// the
				// x axis key
				// 3. compute a TIC value for the retained spectra that will be the y
				// axis value.
				QMap<double, double> rtTicMap;

				int iterationCount = 0;

#pragma omp parallel for
				for(int iter = 0; iter < rtListSize; ++iter)
				{

					if(m_isOperationCancelled)
						continue;

					double rt = rtList.at(iter);

					// This mass spectrum list only holds pointers to msXpSlibmass::MassSpectrum
					// instances
					// allocated on the heap, but it does not own them.
					msXpSlibmass::MassSpectrumList rtMassSpectra(rtHash.values(rt));

					// qDebug() << __FILE__ << __LINE__
					// << "For retention time: " << rt
					// << "number of mass spectra: " << rtMassSpectra.size();

					// We have to remove from all the rtMassSpectra, all the spectra
					// that
					// are not in the drift time range.

					for(int jter = 0; jter < rtMassSpectra.size(); ++jter)
					{
						msXpSlibmass::MassSpectrum *spectrum = rtMassSpectra.at(jter);

						// Old version when msXpSlibmass::MassSpectrum did  not have m_rt/m_dt member
						// variables.
						// double dt = dtHash->key(spectrum);

						if(spectrum->dt() < rangeStart || spectrum->dt() > rangeEnd)
						{
							// Remove the mass spectrum from the rtMassSpectra because
							// it does
							// not have the proper dt value.
							rtMassSpectra.removeAt(jter);
							--jter;
							continue;
						}
					}

					// At this point, the rtMassSpectra only contain spectra that match
					// both the retention time range and the drift time range.

					for(int kter = 0; kter < rtMassSpectra.size(); ++kter)
					{
						double newTic = 0.0;
						if(mzIntegration)
							newTic = rtMassSpectra.at(kter)->tic(mzStart, mzEnd);
						else
							newTic = rtMassSpectra.at(kter)->tic();

						// Had another spectrum by the same rt already been iterated in?
						double oldTic = rtTicMap.value(rt, -1000);

						if(oldTic != -1000)
						{
							// A rt by the same value alreay existed, update the tic
							// value and
							// insert it, that will erase the old value, because the map
							// is
							// NOT a multimap.
							newTic += oldTic;
						}

#pragma omp critical
						rtTicMap.insert(rt, newTic);
					}
					// End of for(int jter = 0; jter < massSpectra.size(); ++jter)

					QString msg = msgSkeleton +
						QString("-- rt %1 with %2 spectra")
						.arg(rt, 0, 'f', 6)
						.arg(rtMassSpectra.size());

#pragma omp critical
					{
						++iterationCount;

						// The label and progress bar widgets were setup by the master
						// thread,
						// so we can only modify them (actually emitting signals) from
						// that
						// same master thread:
						if(omp_get_thread_num() == 0)
							updateProgressFeedback(msg, iterationCount);
					}
				}
				// End of for(int iter = 0; iter < rtList.size(); ++iter)

				// The two commands below work because QMap is ordered along the key and
				// returns the keys in ascending order and the values in ascending order
				// *of the keys*.
				keyVector = QVector<double>::fromList(rtTicMap.keys());
				valVector = QVector<double>::fromList(rtTicMap.values());
			}
			// End of
			// else // that is, !if(massSpecDataSet->isStreamed())

			double end = omp_get_wtime();

			QString msg = QString("Computed TIC chromatogram in %1 s").arg(end - start);

			// Show 100% completion of the work.
			setupProgressFeedback(msg, 0, 100, 100);

			// Hide the cancel button.
			mp_statusBarCancelPushButton->setVisible(false);

			mainWindow->logConsoleMessage(msg);

			// qDebug() << __FILE__ << __LINE__ << msg;

			// In some seconds, hide the progress bar and set the label to an empty
			// text.
			QTimer::singleShot(3000, this, [=]() {
					this->updateProgressFeedback("", 0);
					this->mp_statusBarProgressBar->setVisible(false);
					});

			// Sanity check

			if(keyVector.size() != valVector.size())
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// We may have gotten here because the operation was cancelled. In any
			// case, reset the boolean value.

			m_isOperationCancelled = false;

			// We need to plot that stuff into a new widget.

			// Create a text item to indicate the rt range.

			QString text = QString("%1\n%2 - drift time range : [%3-%4] ")
				.arg(massSpecDataSet->fileName())
				.arg("XIC chromatogram")
				.arg(rangeStart, 0, 'f', 6)
				.arg(rangeEnd, 0, 'f', 6);

			// Allocate the new plot widget in which the data are to be plot.
			TicChromPlotWidget *widget = static_cast<TicChromPlotWidget *>(
					addPlotWidget(keyVector, valVector, text, massSpecDataSet));

			// The history that we get as a parameter needs to be appended to the new
			// widget's history so that we have a full account of the history.

			widget->m_history.copyHistory(history, true /*remove duplicates*/);
			mainWindow->logConsoleMessage(widget->m_history.asText(
						"Plot widget history: \n", "End plot widget history:\n"));
			// And now only the innermost ranges of the whole history.
			mainWindow->logConsoleMessage(widget->m_history.innermostRangesAsText());

			// The plot widget reacts to the Qt::Key_Delete key by calling a function
			// that emits a set of signals, of which the destroyPlotWidget signal
			// actually indirectly triggers the destruction of the plot widget. This
			// signal is
			// now connected to the main window where a function will call for the
			// destruction of all the target plot widgets, that is all the plot
			// widgets that were created as the result of an integration computation
			// of the plot widget currently being destroyed.
			connect(widget, &AbstractPlotWidget::destroyPlotWidget, mainWindow,
					&MainWindow::destroyAlsoAllTargetPlotWidgets);

			connect(widget, &AbstractPlotWidget::toggleMultiGraphGraphSignal,
					mainWindow, &MainWindow::toggleMultiGraphGraph);

			connect(widget, &TicChromPlotWidget::newMassSpectrumByRt,
					mainWindow->mp_massSpecWnd, &MassSpecWnd::newMassSpectrumByRt);

			connect(widget, &TicChromPlotWidget::newDriftSpectrumByRt,
					mainWindow->mp_driftSpecWnd, &DriftSpecWnd::newDriftSpectrumByRt);

			// Finally make the mapping between the tic chrom window plot widget and
			// this new mass spec window plot widget.

			AbstractPlotWidget *senderPlotWidget =
				dynamic_cast<AbstractPlotWidget *>(QObject::sender());
			if(senderPlotWidget == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			mainWindow->m_dataPlotWidgetRelationer.newRelation(
					massSpecDataSet, senderPlotWidget, widget,
					nullptr /* open spectra dialog list widget item */);

			// Ensure that there is a connection between  this plot widget and this
			// containing window, such that if there is a requirement to replot, the
			// widget gets the signal.
			connect(this, &AbstractMultiPlotWnd::replotWithAxisRange, widget,
					&AbstractPlotWidget::replotWithAxisRange);

			// Make sure we get the signal requiring to redraw the background of the
			// widget, depending on it being focused or not.
			connect(this, &AbstractMultiPlotWnd::redrawPlotBackground, widget,
					&AbstractPlotWidget::redrawPlotBackground);
		}


	// We are not going to produce any graph, we are integrating the data to a
	// single TIC intensity value.
	void 
		TicChromWnd::ticIntensity(MassSpecDataSet *massSpecDataSet,
				double *intensity, IntegrationType integrationType, History history)
		{
			if(massSpecDataSet == Q_NULLPTR)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
			if(intensity == Q_NULLPTR)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			double rangeStart;
			double rangeEnd;

			if(!history.innermostRange(integrationType, &rangeStart, &rangeEnd))
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			MainWindow *mainWindow = static_cast<MainWindow *>(parent());
			QString msgSkeleton =
				QString("Computing TIC intensity between rt %1 and rt %2")
				.arg(rangeStart, 0, 'f', 6)
				.arg(rangeEnd, 0, 'f', 6);
			mainWindow->logConsoleMessage(msgSkeleton);

			// Make sure the cancel button is visible.
			mp_statusBarCancelPushButton->setVisible(true);

			double start = omp_get_wtime();

			if(massSpecDataSet->isStreamed())
			{
				streamedTicIntensity(massSpecDataSet, 
						intensity, integrationType, history);
			}
			else // that is, !if(massSpecDataSet->isStreamed())
			{

				// RT integration.
				// ===============
				bool rtIntegration = false;
				double rtStart = -1;
				double rtEnd = -1;

				// If the history is comprised of many successive integrations on the
				// basis of rt value ranges, then we should make sure we do take into
				// account the more internal rt value range.
				rtIntegration = history.innermostRtRange(&rtStart, &rtEnd);
				if(rtIntegration)
				{
					mainWindow->logConsoleMessage(
							"rt integration accounted for: \n" +
							QString("[%1-%2]\n").arg(rtStart).arg(rtEnd));
				}

				// MZ integration.
				// ===============
				bool mzIntegration = false;
				double mzStart = -1;
				double mzEnd = -1;

				// If the history is comprised of many successive integrations on the
				// basis of mz value ranges, then we should make sure we do take into
				// account the more internal mz value range.
				mzIntegration = history.innermostMzRange(&mzStart, &mzEnd);
				if(mzIntegration)
				{
					mainWindow->logConsoleMessage(
							"mz integration accounted for: \n" +
							QString("[%1-%2]\n").arg(mzStart).arg(mzEnd));
				}

				// DT integration.
				// ===============
				bool dtIntegration = false;
				double dtStart = -1;
				double dtEnd = -1;

				// If the history is comprised of many successive integrations on the
				// basis of dt value ranges, then we should make sure we do take into
				// account the more internal dt value range.
				dtIntegration = history.innermostDtRange(&dtStart, &dtEnd);
				if(dtIntegration)
				{
					mainWindow->logConsoleMessage(
							"dt integration accounted for: \n" +
							QString("[%1-%2]\n").arg(dtStart).arg(dtEnd));
				}

				// Get handles to the original data in the calling entity.
				const msXpSlibmass::MsMultiHash &rtHash = massSpecDataSet->rtHash();

				// One way to start the computation is simply to define which retention
				// times should go in the Tic intensity calculation.

				QList<double> rtList;
				if(rtIntegration)
					rtHash.rangedKeys(rtList, rtStart, rtEnd);
				else
					rtList = rtHash.uniqueKeys();

				qSort(rtList.begin(), rtList.end());

				int rtListSize = rtList.size();

				// qDebug() << __FILE__ << __LINE__
				// << "Number of unique retention time keys: " << rtListSize
				// << "from scan start time "<< rtList.first()
				// << "to scan start time" << rtList.last();

				setupProgressFeedback(msgSkeleton, 1, rtListSize);

				// For each retention time, we'll want to
				// 1. get the corresponding mass spectra
				// 2. only retain the spectra that match the drift time (if applicable)
				// 3. compute a TIC value for the retained spectra that is summed 
				// to all the preceding ones, to finally yield a total Tic intensity
				// value (that actually is simply an integration).

				double localIntensity = 0.0;
				int iterationCount = 0;

#pragma omp parallel for
				for(int iter = 0; iter < rtListSize; ++iter)
				{

					if(m_isOperationCancelled)
						continue;

					double rt = rtList.at(iter);

					msXpSlibmass::MassSpectrumList rtMassSpectra(rtHash.values(rt));

					if(dtIntegration)
					{

						// We have to remove from all the rtMassSpectra, all the spectra
						// that are not in the drift time range.

						for(int jter = 0; jter < rtMassSpectra.size(); ++jter)
						{
							msXpSlibmass::MassSpectrum *spectrum = rtMassSpectra.at(jter);

							// Old version when msXpSlibmass::MassSpectrum did not have m_rt/m_dt
							// member
							// variables.
							// double dt = dtHash->key(spectrum);

							if(spectrum->dt() < dtStart || spectrum->dt() > dtEnd)
							{
								// Remove the mass spectrum from the rtMassSpectra
								// because it does
								// not have the proper dt value.
								rtMassSpectra.removeAt(jter);
								--jter;
								continue;
							}
						}
					}

					// At this point, the rtMassSpectra only contain spectra that match
					// both the retention time range and the drift time range.

					for(int kter = 0; kter < rtMassSpectra.size(); ++kter)
					{
						double newTic = 0.0;

						if(mzIntegration)
							newTic = rtMassSpectra.at(kter)->tic(mzStart, mzEnd);
						else
							newTic = rtMassSpectra.at(kter)->tic();

#pragma omp critical
						localIntensity += newTic;
					}
					// End of for(int kter = 0; kter < massSpectra.size(); ++kter)

					QString msg = msgSkeleton +
						QString("-- rt %1 with %2 spectra")
						.arg(rt, 0, 'f', 6)
						.arg(rtMassSpectra.size());

#pragma omp critical
					{
						++iterationCount;

						// The label and progress bar widgets were setup by the master
						// thread, so we can only modify them (actually emitting signals)
						// from that same master thread:

						if(omp_get_thread_num() == 0)
							updateProgressFeedback(msg, iterationCount);
					}
				}
				// End of
				// #pragma omp parallel for
				// for(int iter = 0; iter < rtList.size(); ++iter)

				// Finally, update the value passed to the function:
				*intensity = localIntensity;
			}
			// End of
			// else // that is, !if(massSpecDataSet->isStreamed())

			double end = omp_get_wtime();

			double duration = end - start;

			QString msg = QString("Took %1 s to compute TIC intensity: %2\n")
				.arg(duration)
				.arg(*intensity, 0, 'g', 3);

			// Show 100% completion of the work.
			setupProgressFeedback(msg, 0, 100, 100);

			// Hide the cancel button.
			mp_statusBarCancelPushButton->setVisible(false);

			mainWindow->logConsoleMessage(msg);

			// qDebug() << __FILE__ << __LINE__ << msg;

			// Pretty quickly hide the progress bar and set the label to an empty
			// text so that the user can see the TIC I value on the statusbar.

			// The idea is that if the computation is very quick, then the message
			// should vanish almost immediately. Otherwise let it there longer to show
			// to the user that the computation is computing-intensive.

			// For every second of integration, show the message 20 ms. But do not
			// exceed 3000 ms.
			int timeOut = duration * 20; 
			if(timeOut > 3000)
				timeOut = 3000;

			QTimer::singleShot(timeOut, this, [=]() {
					this->updateProgressFeedback("", 0);
					this->mp_statusBarProgressBar->setVisible(false);
					});

			// We may have gotten here because the operation was cancelled. In any
			// case, reset the boolean value.

			m_isOperationCancelled = false;
		}

	void
		TicChromWnd::newStreamedTicChromatogram(MassSpecDataSet *massSpecDataSet,
				IntegrationType  integrationType,
				QVector<double> *keyVector,
				QVector<double> *valVector,
				History history, QColor color)
		{
			// Depending on the format of the data file from which the data were
			// loaded in stream mode, it will be possible or not to perform the
			// required integration.

			if(massSpecDataSet->fileFormat() ==
					MassSpecDataFileFormat::MASS_SPEC_DATA_FILE_FORMAT_SQLITE3)
			{
				MassSpecDataFileLoaderSqlite3 loader(massSpecDataSet->fileName());

				// Make sure we connect the feedback signal/slot
				connect(&loader,
						&MassSpecDataFileLoaderSqlite3::setupProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::setupProgressFeedback);

				connect(&loader,
						&MassSpecDataFileLoaderSqlite3::updateProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::updateProgressFeedback);

				connect(this, &AbstractMultiPlotWnd::cancelOperationSignal, &loader,
						&MassSpecDataFileLoaderSqlite3::cancelOperation);

				int res = loader.streamedIntegration(
						history, IntegrationType::ANY_TO_RT, keyVector, valVector);

				if(res == -1)
					qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
			}
			else if(massSpecDataSet->fileFormat() ==
					MassSpecDataFileFormat::MASS_SPEC_DATA_FILE_FORMAT_MZML)
			{
				MassSpecDataFileLoaderPwiz loader(massSpecDataSet->fileName());

				// Make sure we connect the feedback signal/slot
				connect(&loader,
						&MassSpecDataFileLoaderPwiz::setupProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::setupProgressFeedback);

				connect(&loader,
						&MassSpecDataFileLoaderPwiz::updateProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::updateProgressFeedback);

				connect(this, &AbstractMultiPlotWnd::cancelOperationSignal,
						&loader, &MassSpecDataFileLoaderPwiz::cancelOperation);

				int res = loader.streamedIntegration(
						history, IntegrationType::ANY_TO_RT, keyVector, valVector);

				if(res == -1)
					qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
			}
			else
			{
				QMessageBox msgBox;
				msgBox.setText(
						"The operation cannot be carried over because the "
						"data were loaded in stream and not from a database file.");

				msgBox.setStandardButtons(QMessageBox::Ok);
				msgBox.exec();

				// Hide the cancel button.
				mp_statusBarCancelPushButton->setVisible(false);

				return;
			}
		}


	void
		TicChromWnd::streamedTicIntensity(MassSpecDataSet *massSpecDataSet,
				double *intensity,IntegrationType integrationType, History history)
		{
			if(intensity == Q_NULLPTR)
				qFatal("Fatal error at %s@%d. Program aborted.",
						__FILE__, __LINE__);

			// Depending on the format of the data file from which the data were
			// loaded in stream mode, it will be possible or not to perform the
			// required integration.

			if(massSpecDataSet->fileFormat() ==
					MassSpecDataFileFormat::MASS_SPEC_DATA_FILE_FORMAT_SQLITE3)
			{
				MassSpecDataFileLoaderSqlite3 loader(massSpecDataSet->fileName());

				// Make sure we connect the feedback signal/slot
				connect(&loader,
						&MassSpecDataFileLoaderSqlite3::setupProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::setupProgressFeedback);

				connect(&loader,
						&MassSpecDataFileLoaderSqlite3::updateProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::updateProgressFeedback);

				connect(this, &AbstractMultiPlotWnd::cancelOperationSignal, &loader,
						&MassSpecDataFileLoaderSqlite3::cancelOperation);

				int res = loader.streamedTicIntensity(history, intensity);

				if(res == -1)
					qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
			}
			else if(massSpecDataSet->fileFormat() ==
					MassSpecDataFileFormat::MASS_SPEC_DATA_FILE_FORMAT_MZML)
			{
				MassSpecDataFileLoaderPwiz loader(massSpecDataSet->fileName());

				// Make sure we connect the feedback signal/slot
				connect(&loader,
						&MassSpecDataFileLoaderPwiz::setupProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::setupProgressFeedback);

				connect(&loader,
						&MassSpecDataFileLoaderPwiz::updateProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::updateProgressFeedback);

				connect(this, &AbstractMultiPlotWnd::cancelOperationSignal,
						&loader, &MassSpecDataFileLoaderPwiz::cancelOperation);

				int res = loader.streamedTicIntensity(history, intensity);

				if(res == -1)
					qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
			}
			else if(massSpecDataSet->fileFormat() ==
					MassSpecDataFileFormat::MASS_SPEC_DATA_FILE_FORMAT_DX)
			{
				MassSpecDataFileLoaderDx loader(massSpecDataSet->fileName());

				// Make sure we connect the feedback signal/slot
				connect(&loader,
						&MassSpecDataFileLoaderDx::setupProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::setupProgressFeedback);

				connect(&loader,
						&MassSpecDataFileLoaderDx::updateProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::updateProgressFeedback);


				connect(this, &AbstractMultiPlotWnd::cancelOperationSignal,
						&loader, &MassSpecDataFileLoaderDx::cancelOperation);

				int res = loader.streamedIntegration(history, 
						Q_NULLPTR /* keyVector */, Q_NULLPTR /* valVector */, intensity);

				if(res == -1)
					qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
			}
			else if(massSpecDataSet->fileFormat() ==
					MassSpecDataFileFormat::MASS_SPEC_DATA_FILE_FORMAT_MS1)
			{
				MassSpecDataFileLoaderMs1 loader(massSpecDataSet->fileName());

				// Make sure we connect the feedback signal/slot
				connect(&loader,
						&MassSpecDataFileLoaderMs1::setupProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::setupProgressFeedback);

				connect(&loader,
						&MassSpecDataFileLoaderMs1::updateProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::updateProgressFeedback);

				connect(this, &AbstractMultiPlotWnd::cancelOperationSignal,
						&loader, &MassSpecDataFileLoaderMs1::cancelOperation);

				int res = loader.streamedIntegration(
						history, Q_NULLPTR /* keyVector */, Q_NULLPTR /* valVector */, intensity);

				if(res == -1)
					qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
			}
			else
			{
				QMessageBox msgBox;
				msgBox.setText(
						"The operation cannot be carried over because the "
						"data were loaded in stream and not from a database file.");

				msgBox.setStandardButtons(QMessageBox::Ok);
				msgBox.exec();

				// Hide the cancel button.
				mp_statusBarCancelPushButton->setVisible(false);

				return;
			}
		}	

	void
		TicChromWnd::toggleMultiGraphGraph()
		{
			AbstractPlotWidget *ticChromPlotWidget =
				dynamic_cast<AbstractPlotWidget *>(QObject::sender());
			QCPGraph *graph = m_plotWidgetGraphMap.value(ticChromPlotWidget);

			if(graph == nullptr)
				return;

			// First toggle visibility of this plot widget graph.
			if(graph->visible())
				graph->setVisible(false);
			else
				graph->setVisible(true);

			// And now ask for a replot of the multi-graph plot widget.
			mp_multiGraphPlotWidget->replot();

			// But we know that each tic chrom plot widget might have a number of mass
			// spec window plot widget that derive from it by way of mass spectrum
			// combination or also any number of drift spectra by the same integration
			// mechanism. All these  plot widgets must be have
			// their visibility toggled also. That is, we need to get a recursive
			// search list of all the plot widget targets that were directly or
			// indirectly generated from this ticChromPlotWidget.

			MainWindow *mainWindow = static_cast<MainWindow *>(parent());

			QList<AbstractPlotWidget *> plotWidgetList;
			mainWindow->m_dataPlotWidgetRelationer.targetsRecursive(ticChromPlotWidget,
					&plotWidgetList);

			for(int iter = 0; iter < plotWidgetList.size(); ++iter)
			{
				plotWidgetList.at(iter)->toggleMultiGraphGraph();
			}
		}


	void
		TicChromWnd::clearPlots()
		{
			AbstractMultiPlotWnd::clearPlots();
		}


	void
		TicChromWnd::recordAnalysisStanza(const QString &stanza, const QColor &color)
		{
			// The record might go to a file or to the console or both.

			if((mp_analysisPreferences->m_recordTarget &
						RecordTarget::RECORD_TO_FILE) == RecordTarget::RECORD_TO_FILE)
			{
				if(mp_analysisFile != nullptr)
				{

					// Write the stanza, that was crafted by the calling plot widget to
					// the file.

					if(!mp_analysisFile->open(QIODevice::Append))
					{
						statusBar()->showMessage(
								QString("Could not record the step because "
									"the file could not be opened."),
								4000);
					}
					else
					{
						mp_analysisFile->write(stanza.toLatin1());
						// Force writing because we may want to have tail -f work fine
						// on the
						// file, and see modifications live to change fiels in a text
						// editor.
						mp_analysisFile->flush();
						mp_analysisFile->close();
					}
				}
				else
				{
					qDebug() << __FILE__ << __LINE__
						<< "The mp_analysisFile pointer is nullptr.";

					statusBar()->showMessage(
							QString("Could not record the analysis step to file. "
								"Please define a file to write the data to."),
							4000);
				}
			}

			// Also, if recording to the console is asked for, then do that also.
			if((mp_analysisPreferences->m_recordTarget &
						RecordTarget::RECORD_TO_CONSOLE) == RecordTarget::RECORD_TO_CONSOLE)
			{
				MainWindow *mainWindow = static_cast<MainWindow *>(parent());
				mainWindow->logConsoleMessage(stanza, color);
			}

			// Also, if recording to the clipboard is asked for, then do that also.
			if((mp_analysisPreferences->m_recordTarget &
						RecordTarget::RECORD_TO_CLIPBOARD) == RecordTarget::RECORD_TO_CLIPBOARD)
			{
				QClipboard *clipboard = QApplication::clipboard();
				clipboard->setText(clipboard->text() + stanza, QClipboard::Clipboard);
			}

			return;
		}


} // namespace msXpSmineXpert
