/*  BEGIN software license
 *
 *  msXpertSuite - mass spectrometry software suite
 *  -----------------------------------------------
 *  Copyright(C) 2009, 2017 Filippo Rusconi
 *
 *  http://www.msxpertsuite.org
 *
 *  This file is part of the msXpertSuite project.
 *
 *  The msXpertSuite project is the successor of the massXpert project. This
 *  project now includes various independent modules:
 *  
 *  - massXpert, model polymer chemistries and simulate mass spectrometric data;
 *  - mineXpert, a powerful TIC chromatogram/mass spectrum viewer/miner;
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * END software license
 */



/////////////////////// Std includes


/////////////////////// Qt includes
#include <QFileInfo>
#include <QList>
#include <QStringList>
#include <QDebug>
#include <qmath.h>


/////////////////////// Local includes
#include "CommandLineRunner.hpp"
#include "MassSpecDataFileLoaderSqlite3.hpp"
#include "MassSpecDataFileFormatAnalyzer.hpp"
#include "MassSpecDataFileLoaderXy.hpp"
#include "FileFormatConverter.hpp"

#include "globals/globals.hpp"
#include "globals.hpp"


namespace msXpSmineXpert
{


	//! Construct an empy CommandLineRunner instance
	CommandLineRunner::CommandLineRunner()
	{
	}


	//! Destruct \c this CommandLineRunner instance 
	CommandLineRunner::~CommandLineRunner()
	{
	}


	//! Tell which input file names match existing files on disk.
	/*!

		When the parsing of the command line is performed, the input file names are
		available in a list. Iterate in the list and for each file name determine if
		it exists on disk. If a file name matches a file on disk, append its file name to the returned string list.

		\return A string list containing all the files names that matched a
		corresponding existing file on disk.

		\sa allInputFilesExist()

*/
	QStringList
		CommandLineRunner::existingInputFiles()
		{
			QStringList existingFiles;

			for(int iter = 0; iter < m_inFileNames.size(); ++iter)
			{
				m_currentInputFileName = m_inFileNames.at(iter);
				QFileInfo fileInfo(m_currentInputFileName);

				if(fileInfo.exists())
					existingFiles.append(m_currentInputFileName);
				else
					msXpS::qStdOut() << __FILE__ << __LINE__ << __FUNCTION__ 
						<< "File" << m_currentInputFileName << "not found";
			}

			return existingFiles;
		}



	//! Tell if all the files in input actually do exist
	/*!

		When the parsing of the command line is performed, the input file names are
		available in a list. Iterate in the list and for each file name determine if
		it exists on disk. If a file name does not match an existing file on disk,
		increment \p *missingFiles. Conversely, if a matching file on disk is found,
		increment \p *existingFiles.

		\param existingFiles pointer to an int where to set the number of existing
		files on disk. Might be nullptr, in which case the counter is not modified.

		\param missingFiles pointer to an int where to set the number of missing
		files on disk. Might be nullptr, in which case the counter is not modified.

		\return true if all the files were found, false otherwise.

		\sa existingInputFiles()
		\sa allInputFilesExist()
		*/
	bool
		CommandLineRunner::allInputFilesExist(int *existingFiles, int *missingFiles)
		{
			// Initialize the variables to meaningful values

			int notFound = 0;
			int found = 0;

			for(int iter = 0; iter < m_inFileNames.size(); ++iter)
			{
				QFileInfo fileInfo(m_inFileNames.at(iter));

				if(!fileInfo.exists())
				{
					++notFound;
				}
				else
				{
					++found;
				}
			}

			if(existingFiles != nullptr)
				*existingFiles = found;

			if(missingFiles != nullptr)
				*missingFiles = notFound;

			return (m_inFileNames.size() == found);
		}


	//! Determines which input file exist or do not
	/*!

		When the parsing of the command line is performed, the input file names are
		available in a list. Iterate in the list and for each file name determine if
		it exists on disk. If a file name does not match an existing file on disk,
		append its file name to \p *missingFiles. Conversely, if a matching file on disk is found,
		append the file name to \p *existingFiles.

		\param existingFiles pointer to a string list where to store the file names
		of the found files.

		\param missingFiles pointer to a string list where to store the file names
		of the not-found files.

		\return true if all the files were found, false otherwise.

		\sa existingInputFiles()
		\sa allInputFilesExist()

*/
	bool
		CommandLineRunner::allInputFilesExist(QStringList *existingFiles,
				QStringList *missingFiles)
		{
			if(existingFiles == nullptr || missingFiles == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			bool allExist = true;

			for(int iter = 0; iter < m_inFileNames.size(); ++iter)
			{
				m_currentInputFileName = m_inFileNames.at(iter);
				QFileInfo fileInfo(m_currentInputFileName);

				if(!fileInfo.exists())
				{
					allExist = false;
					missingFiles->append(m_currentInputFileName);
				}
				else
					existingFiles->append(m_currentInputFileName);
			}

			return allExist;
		}


	//! Tell if the GUI should be run right away.
	/*!

		When parsing the command line, establish if the argument to the program call
		are such that the program's graphical user interface must be run without
		delay. The program might indeed be run in console mode, when a file format
		conversion is run.

		A typical example that makes this function return false is when a file
		format conversion is required. Another typical example is when a list of
		file names is provided with no other argument: this function returns true
		and the GUI load all these files. Finally, if the single argument is the
		file name of a script file, then the program should run the GUI right away
		and start executing the commands in the script file.

		Note that note of the parameters might be nullptr.

		\param existingFiles pointer to a string list in which to store the file
		names of the existing files actually found on disk.

		\param errors pointer to a string in which to describe the errors encountered. 

		\return true if the program can be run through its GUInterface; false
		otherwise.

*/
	bool
		CommandLineRunner::shouldRunGuiRightAway(QStringList *existingFiles,
				QString *errors)
		{
			if(existingFiles == Q_NULLPTR || errors == Q_NULLPTR)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// As soon as we have the conversion flag on, we should do the work, only
			// if the list of files is ok, in input.

			if(m_isFileConversion && !m_inFileNames.isEmpty())
				return false;

			// Make sure we set the existing files to the parameter. The user might
			// have type an erroneous file name and we do not want to fail totally for
			// a single error.
			existingFiles->append(existingInputFiles());

			// At this point, start right away with the GUI.
			qDebug() << __FILE__ << __LINE__ << __FUNCTION__ << "Returning true.";

			return true;
		}


	//! Main acting function that triggers the running of the program.
	/*!

		This functions analyses (using helper functions) the command line parameters
		(if any) and takes action depending on these parameters.

		For example, if a file format conversion is asked for, that conversion is
		run without even starting the Qt application execution loop.

		\param existingFiles pointer to a string list in which to store the file
		names of files effectively found on disk. May not be nullptr.

		\param errors pointer to a string in which to store all the errors
		encountered during the analysis of the command line. May not be nullptr.

		\param shouldRunGui pointer to a boolean value in which to tell if the
		program should run the graphical user interface. May not be nullptr.

		\return 

		- true if the GUI should be run right away
		- true if a script file name was provided
		- the exit status of a conversion run if a conversion run is started
		- false otherwise

		\sa shouldRunGuiRightAway()

*/
	bool
		CommandLineRunner::run(QStringList *existingFiles, QString *errors,
				bool *shouldRunGui)
		{
			if(existingFiles == Q_NULLPTR || errors == Q_NULLPTR || shouldRunGui == Q_NULLPTR)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// Before anything, if the arguments contain a script file name, then
			// return should run gui = true.

			if(!m_scriptFileName.isEmpty())
			{
				qDebug() << __FILE__ << __LINE__ << __FUNCTION__ << "Script file name non-empty.";

				*shouldRunGui = true;
				return true;
			}

			// This command returns false if there was at least one error in the
			// command line parameters that impedes to launch the program. Upon
			// returning, shouldRunGui should have been set to the proper value.

			if(shouldRunGuiRightAway(existingFiles, errors))
			{
				*shouldRunGui = true;
				return true;
			}

			// Is a file conversion asked for?
			if(m_isFileConversion)
			{
				qDebug() << __FILE__ << __LINE__ << __FUNCTION__ << "Conversion asked.";

				return runConversion(errors);
			}

			return false;
		}


	//! Start a file format conversion
	/*!

		mineXpert can run in graphical user interface mode but also without a GUI
		when a file format conversion is asked for. Typically, this is the case when
		the user passes argument to the command line to convert a mzML file into the
		private SQLite3-based file format.

		The actual file format conversion work is performed by the
		FileFormatConverter class.

		\param errors pointer to a string in which to append errors encountered
		during the run of the file format conversion.

		\return true if the conversion was successful; false otherwise.

		\sa FileFormatConverter.

*/
	bool 
		CommandLineRunner::runConversion(QString *errors)
		{
			bool success = true;

			qDebug() << __FILE__ << __LINE__ << "Entering runConversion().";

			// For each input file, create a new file that reproduces the data but in
			// SQLite3 db format.

			FileFormatConverter loader;

			for(int iter = 0; iter < m_inFileNames.size(); ++iter)
			{
				loader.setInFileName(m_inFileNames.at(iter));
				loader.setOutFileName(m_outFileName);

				success = loader.convert(errors);

				if(success)
				{
					qDebug() << __FILE__ << __LINE__
						<< "Conversion for file:" << loader.inFileName() << "success.";
				}
				else
					qDebug() << __FILE__ << __LINE__
						<< "Conversion for file:" << loader.inFileName() << "failure.";
			}

			return success;
		}


	//! Craft a string with all the command line arguments.
	/*!

		\return A string containing all the command line parameters nicely formatted.

*/
	QString
		CommandLineRunner::asText()
		{
			QString text = "input files: " + m_inFileNames.join(" , ") + "\n";
			text += "output file: " + m_outFileName + "\n";
			text += "conversion: " + QString("%1\n")
				.arg(m_isFileConversion ? "true" : "false");
			text += "script file: " + m_scriptFileName + "\n";

			return text;
		}


} // namespace msXpSmineXpert
